# coding: utf8
"""
Wide variety of helper methods for working with iterables.

A kind-of addition to `itertools`.
"""

from __future__ import print_function, unicode_literals, absolute_import, division

import os
from itertools import chain, repeat, islice


__all__ = (
    'window',
    'reversed_blocks',
    'reversed_lines',
    'uniq_g',
    'IterStat',
    'IterMean',
    'chunks_g',
    'next_or_fdefault',
    'iterator_is_over',
)


# Iterate over a 'window' of adjacent elements
# http://stackoverflow.com/questions/6998245/iterate-over-a-window-of-adjacent-elements-in-python
def window(seq, size=2, fill=0, fill_left=False, fill_right=False):
    """ Returns a sliding window (of width n) over data from the iterable:
      s -> (s0,s1,...s[n-1]), (s1,s2,...,sn), ...
    """
    ssize = size - 1
    it = chain(
        repeat(fill, ssize * fill_left),
        iter(seq),
        repeat(fill, ssize * fill_right))
    result = tuple(islice(it, size))
    if len(result) == size:  # `<=` if okay to return seq if len(seq) < size
        yield result
    for elem in it:
        result = result[1:] + (elem,)
        yield result


def pair_window(iterable):
    """
    A more simple version of `window` for size=2.

    >>> list(window([11, 22, 33, 44]))
    [(11, 22), (22, 33), (33, 44)]
    """
    iterable = iter(iterable)
    try:
        prev_value = next(iterable)
    except StopIteration:
        return
    for item in iterable:
        yield prev_value, item
        prev_value = item


def cumsum(iterable):
    """
    Cumulative sum.

    >>> list(cumsum([1.2, 3.4, 5.6]))
    [1.2, 4.6, 10.2]
    """
    iterable = iter(iterable)
    try:
        value = next(iterable)
    except StopIteration:
        return
    yield value
    for item in iterable:
        value += item
        yield value


# ###### Reading files backwards
# http://stackoverflow.com/a/260433/62821

def reversed_blocks(fileobj, blocksize=4096):
    """ Read blocks of file's contents in reverse order.  """
    fileobj.seek(0, os.SEEK_END)
    here = fileobj.tell()
    while here > 0:
        delta = min(blocksize, here)
        fileobj.seek(here - delta, os.SEEK_SET)
        yield fileobj.read(delta)
        here -= delta


def reversed_lines(fileobj, sep=b''):
    """ Read the lines of file in reverse order """
    tail = []           # Tail of the line whose head is not yet read.
    for block in reversed_blocks(fileobj):
        # A line is a list of strings to avoid quadratic concatenation.
        # (And trying to avoid 1-element lists would complicate the code.)
        linelists = [[line] for line in block.splitlines()]
        linelists[-1].extend(tail)
        for linelist in reversed(linelists[1:]):
            yield sep.join(linelist)
        tail = linelists[0]
    if tail:
        yield sep.join(tail)


def uniq_g(lst, key=lambda value: value):
    """
    Get unique elements of an iterable preserving its order and optionally
    determining uniqueness by hash of a key.
    """
    known = set()
    for value in lst:
        value_key = key(value)
        if value_key not in known:
            yield value
            known.add(value_key)


uniq = uniq_g


class IterStat(object):
    """
    Iterative single-pass computing of mean and variance.

    Error is on the rate of 1e-08 for 1e6 values in the range of
    0..1e6, both for mean and for stddev.

    http://www.johndcook.com/standard_deviation.html
    """

    def __init__(self, vals=None, start=0):
        self.start = start
        self.old_mean = None
        self.mean = self.stdx = start
        self.cnt = 0

        if vals:
            for val in vals:
                self.send(val)

    def send(self, val):
        self.cnt += 1
        if self.cnt == 1:
            self.mean = val
        else:
            self.mean = self.mean + (val - self.mean) / float(self.cnt)
            self.stdx = self.stdx + (val - self.old_mean) * (val - self.mean)
        self.old_mean = self.mean

    @property
    def variance(self):
        if self.cnt <= 1:
            return self.start
        return self.stdx / (self.cnt)

    @property
    def std(self):
        from .base import _sqrt
        return _sqrt(self.variance)


def IterMean(iterable, dtype=float):
    """ Mean of an iterable """
    res_sum, cnt = dtype(), dtype()
    for val in iterable:
        res_sum += val
        cnt += 1
    if cnt == 0:  # NOTE.
        try:
            return dtype('nan')
        except Exception:
            return float('nan')
    return res_sum / cnt


def chunks_g(iterable, size):
    """
    Same as 'chunks' but works on any iterable.

    Converts the chunks to tuples for simplicity.

    http://stackoverflow.com/a/8991553
    """
    it = iter(iterable)
    if size <= 0:
        yield it
        return
    while True:
        chunk = tuple(islice(it, size))
        if not chunk:
            return
        yield chunk


def next_or_fdefault(it, default=lambda: None, skip_empty=False):
    """
    `next(it, default_value)` with laziness.

    >>> next_or_fdefault([1], lambda: 1/0)
    1
    >>> next_or_fdefault([], lambda: list(range(2)))
    [0, 1]
    """
    if skip_empty:
        it = (val for val in it if val)
    else:
        it = iter(it)
    try:
        return next(it)
    except StopIteration:
        return default()


def iterator_is_over(it, ret_value=False):
    """
    Try to consume an item from an iterable `it` and return False if it
    succeeded (the item stays consumed).
    """
    try:
        val = next(it)
    except StopIteration:
        if ret_value:
            return True, None
        return True
    else:
        if ret_value:
            return False, val
        return False


def with_last(it):
    """
    Wrap an iterable yielding an `is_last, value`.

    >>> list(with_last([]))
    []
    >>> list(with_last([1]))
    [(True, 1)]
    >>> list(with_last([1, 2]))
    [(False, 1), (True, 2)]
    >>> list(with_last([1, 2, 3]))
    [(False, 1), (False, 2), (True, 3)]
    """
    it = iter(it)
    try:
        prev_value = next(it)
    except StopIteration:
        return
    value = prev_value
    for value in it:
        yield False, prev_value
        prev_value = value
    yield True, value


class NotEnoughItems(Exception):
    """ Pretty much StopIteration but only for handling by user-code """


def prefetch_first(iterable, count=1, require=False):
    """
    Transparent-ish iterable wrapper that obtains the first N items on call.

    Not to be confused with continuous peeking at any point in iteration; that
    might as well be done with `window`.

    Useful for error-catching and in parallelizing.

    :param iterable: ...

    :param count: amount of items to prefetch.

    :param require: if True, raises an NotEnoughItems when it couldn't prefetch
    `count` items.

    Unfortunately, it is not possible to assign a convenience attribute on the
    result (which is a generator object),
    but the current list of the prefetched items in python3 can be found at

        iterable.gi_frame.f_locals['prefetched']

    >>> def verbose_iter(num):
    ...     for idx in range(num):
    ...         print("verbose_iter", num, idx)
    ...         yield idx
    ...
    >>> iterable = verbose_iter(2)
    >>> list(iterable)
    verbose_iter 2 0
    verbose_iter 2 1
    [0, 1]
    >>> iterable = verbose_iter(3)
    >>> iterable = prefetch_first(iterable, count=2)
    verbose_iter 3 0
    verbose_iter 3 1
    >>> iterable.gi_frame.f_locals['prefetched']
    [0, 1]
    >>> list(iterable)
    verbose_iter 3 2
    [0, 1, 2]
    >>> list(prefetch_first(verbose_iter(1), count=10))
    verbose_iter 1 0
    [0]
    >>> list(prefetch_first(verbose_iter(0), count=10))
    []
    >>> list(prefetch_first(verbose_iter(0), count=0))
    []
    >>> iterable = prefetch_first(verbose_iter(2), count=0)
    >>> list(iterable)
    verbose_iter 2 0
    verbose_iter 2 1
    [0, 1]
    """
    iterable = iter(iterable)
    prefetched = []
    for _ in range(count):
        try:
            prefetched.append(next(iterable))
        except StopIteration:
            if require:
                raise NotEnoughItems(
                    "Could not prefetch the requested amount of items",
                    dict(requested=count, found=len(prefetched), data=prefetched))

    # pylint: disable=dangerous-default-value
    def gen(iterable=iterable, prefetched=prefetched):
        if prefetched:
            for item in prefetched[:-1]:
                yield item
            # A few extra tricks to drop the references to prefetched items without
            # impacting performance too much.
            item = prefetched[-1]
            prefetched[:] = []
            yield item
        for item in iterable:
            yield item

    return gen()


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# coding: utf8
# pylint: disable=useless-import-alias
"""
Monkey-patching of various things
"""

from __future__ import division, absolute_import, print_function, unicode_literals


__all__ = (
    'use_cdecimal',
    'use_exc_ipdb',
    'use_exc_log',
    'use_colorer',
)


def use_cdecimal():
    """ Do a hack-in replacement of `decimal` with `cdecimal`.
    Should be done before importing other modules.

    Also see
    http://adamj.eu/tech/2015/06/06/swapping-decimal-for-cdecimal-on-python-2/
    for a possibly more reliable way.
    """
    import sys
    import decimal  # maybe not needed
    import cdecimal  # pylint: disable=import-error
    sys.modules['decimal'] = cdecimal


def use_exc_ipdb():
    """ Set unhandled exception handler to automatically start ipdb """
    import pyaux.exc_ipdb as exc_ipdb
    exc_ipdb.init()


def use_exc_log():
    """ Set unhandled exception handler to verbosely log the exception """
    import pyaux.exc_log as exc_log
    exc_log.init()


def use_colorer():
    """ Wrap logging's StreamHandler.emit to add colors to the logged
      messages based on log level """
    # TODO: make a ColorerHandlerMixin version
    import pyaux.Colorer as Colorer
    Colorer.init()



# coding: utf8
"""
Automatically start the debugger on an exception.

Replaces sys.excepthook on `init()`.
Can be included in 'sitecustomize.py'

src: http://code.activestate.com/recipes/65287-automatically-start-the-debugger-on-an-exception/
"""

import sys


def info(type, value, tb):
    if hasattr(sys, 'ps1') or not sys.stderr.isatty():
        # we are in interactive mode or we don't have a tty-like
        # device, so we call the default hook
        sys.__excepthook__(type, value, tb)
    else:
        import traceback, ipdb
        # we are NOT in interactive mode, print the exception...
        traceback.print_exception(type, value, tb)
        print
        # ...then start the debugger in post-mortem mode.
        ipdb.pm()


def init():
    sys.excepthook = info



# coding: utf8
# pylint: disable=invalid-name,unused-import,import-error
"""
Some pieces of `six`-like code relevant to this package.
"""

from __future__ import division, absolute_import, print_function, unicode_literals

import sys


PY_3 = sys.version_info >= (3,)
PY3 = PY_3


if PY_3:
    import urllib.parse as urllib_parse

    izip = zip
    unicode = str
    text_type = str
    xrange = range

    def reraise(exc_type, exc_value, exc_traceback):  # pylint: disable=unused-argument
        try:
            raise exc_value.with_traceback(exc_traceback)
        finally:
            exc_value = None
            exc_traceback = None

else:

    import urlparse as urllib_parse

    from itertools import izip  # pylint: disable=no-name-in-module

    unicode = unicode
    text_type = unicode
    xrange = xrange

    # Would be a syntax error in py3, have to exec-wrap it.
    exec((  # pylint: disable=exec-used
        "def reraise(exc_type, exc_value, exc_traceback):\n"
        "    try:\n"
        "        raise exc_type, exc_value, exc_traceback\n"
        "    finally:\n"
        "        exc_traceback = None\n"
    ))



"""
...

NOTE: python3.5+ only (only tested on python3.7).
"""
# pylint: disable=no-else-return
# pylint: disable=too-many-branches
# pylint: disable=too-many-statements
# pylint: disable=fixme

import copy
import re
# pylint: disable=no-name-in-module
from sre_constants import (
    LITERAL, RANGE, IN, NOT_LITERAL, NEGATE, MIN_REPEAT, MAX_REPEAT, MAXREPEAT,
    ANY, AT, AT_BEGINNING, AT_END, SUBPATTERN, BRANCH,
)
import sre_parse
import sre_compile


UNESCAPE = {val: key for key, val in sre_parse.ESCAPES.items()}
UNCATEGORIES = {
    ((val[0], tuple(val[1])) if isinstance(val, tuple) and len(val) == 2 and isinstance(val[1], list) else val): key
    for key, val in sre_parse.CATEGORIES.items()}


def _ensure_grouped(pattern, source=None):
    """
    ...

    >>> _ensure_grouped(pattern='zxcv', source=[(LITERAL, 122), (LITERAL, 120), (LITERAL, 99), (LITERAL, 118)])
    '(?:zxcv)'
    >>> _ensure_grouped(pattern='(?:zxcv)', source=[(LITERAL, 122), (LITERAL, 120), (LITERAL, 99), (LITERAL, 118)])
    '(?:zxcv)'
    >>> _ensure_grouped(pattern='[abc]', source=[(IN, [(LITERAL, 97), (LITERAL, 98), (LITERAL, 99)])])
    '[abc]'
    """
    if pattern.startswith('(') or pattern.startswith('['):
        return pattern
    if len(pattern) == 1:
        return pattern
    # TODO: e.g. '\r*'
    # if not source: source = sre_parse.parse(pattern)
    # if len(source.data) == 1: return pattern
    return '(?:{})'.format(pattern)


def _flags_to_list(flags):
    result = []
    if not flags:
        return result
    for key, value in sre_parse.FLAGS.items():
        if flags & value:
            result.append(key)
    return result


# TODO: re.VERBOSE autoindented version.
def rast_to_pattern(rast, _parent_type=None, **kwargs):
    r"""
    Inverse of `sre_parse._parse`.

    Partial implementation (to be completed as needed).

    >>> rex = r'[abc]?(?:^|a{4,}?)[^IO](b\ (?P<zxcv>c|d)+)*($|...|[0-9×][^a-z])(?P<qwer>a-imsx:zxcv)(?a-xsmi:zxcv)'
    >>> rast_to_pattern(sre_parse.parse(rex))
    '[abc]?(?:^|a{4,}?)[^IO](b\\ (?P<zxcv>[cd])+)*($|...|[0-9×][^a-z])(?P<qwer>a\\-imsx:zxcv)(?a-imsx:zxcv)'
    """
    _chr = chr
    # TODO: bytes-regex support: `if not kwargs.get(flags, 0) & re.UNICODE: _chr = lambda num: bytes([num])`

    # # # Reference glue;
    # def _parse(source, state, verbose, nested, first=False):
    # # parse a simple pattern
    # subpattern = SubPattern(state)

    # # # Reference glue;
    # # precompute constants into local variables
    # subpatternappend = subpattern.append
    # sourceget = source.get
    # sourcematch = source.match
    # _len = len
    # _ord = ord

    if isinstance(rast, sre_parse.SubPattern):

        kwargs['flags'] = kwargs.get('flags', 0) | rast.pattern.flags

        if kwargs.get('group_to_name') is None:
            kwargs['group_to_name'] = {
                val: key
                for key, val in rast.pattern.groupdict.items()}

        # Tricky point: support 'branch inside subpattern does not require extra parentheses',
        # AST goes like `SUBPATTERN -> SubPattern -> BRANCH`.
        # Thus, in a general-ish case, pass parent through a single-element SubPattern.
        if len(rast.data) == 1:
            kwargs['_parent_type'] = _parent_type

        return ''.join(
            rast_to_pattern(child, **kwargs)
            for child in rast)

    # # Reference glue:
    # while True:
    #     this = source.next
    #     if this is None:
    #         break # end of pattern
    #     if this in "|)":
    #         break # end of subpattern
    #     sourceget()

    # # Not supported in this function (but could be):
    # # Reference:
    #     if verbose:
    #         # skip whitespace and comments
    #         if this in WHITESPACE:
    #             continue
    #         if this == "#":
    #             while True:
    #                 this = sourceget()
    #                 if this is None or this == "\n":
    #                     break
    #             continue

    elif isinstance(rast, tuple) and len(rast) == 2:
        item_type, item_value = rast
        # For recursion.
        kwargs['_parent_type'] = item_type

        # # Reference:
        # if this[0] == "\\":
        #     code = _escape(source, this, state)
        #     subpatternappend(code)
        # # Reference:
        # elif this not in SPECIAL_CHARS:
        #     subpatternappend((LITERAL, _ord(this)))
        if item_type is LITERAL and isinstance(item_value, int):
            # XXXX: re.escape (re._special_chars_map) does not quite match
            # the sre_parse.ESCAPES. Might prefer to use the latter here.
            # The `re.escape` tends to escape more than needed, e.g. ' ' or '-' in all cases.
            return re.escape(_chr(item_value))

        # # Reference:
        # elif this == "[":
        #     here = source.tell() - 1
        #     # character set
        #     set = []
        #     setappend = set.append
        #     # if sourcematch(":"):
        #         # pass # handle character classes
        #     if source.next == '[':
        #         import warnings
        #         warnings.warn(
        #             'Possible nested set at position %d' % source.tell(),
        #             FutureWarning, stacklevel=nested + 6
        #         )
        #     negate = sourcematch("^")
        #     # check remaining characters
        #     while True:
        #         this = sourceget()
        #         if this is None:
        #             raise source.error("unterminated character set",
        #                                source.tell() - here)
        #         if this == "]" and set:
        #             break
        #         elif this[0] == "\\":
        #             code1 = _class_escape(source, this)
        #         else:
        #             if set and this in '-&~|' and source.next == this:
        #                 import warnings
        #                 warnings.warn(
        #                     'Possible set %s at position %d' % (
        #                         'difference' if this == '-' else
        #                         'intersection' if this == '&' else
        #                         'symmetric difference' if this == '~' else
        #                         'union',
        #                         source.tell() - 1),
        #                     FutureWarning, stacklevel=nested + 6
        #                 )
        #             code1 = LITERAL, _ord(this)
        #         if sourcematch("-"):
        #             # potential range
        #             that = sourceget()
        #             if that is None:
        #                 raise source.error("unterminated character set",
        #                                    source.tell() - here)
        #             if that == "]":
        #                 if code1[0] is IN:
        #                     code1 = code1[1][0]
        #                 setappend(code1)
        #                 setappend((LITERAL, _ord("-")))
        #                 break
        #             if that[0] == "\\":
        #                 code2 = _class_escape(source, that)
        #             else:
        #                 if that == '-':
        #                     import warnings
        #                     warnings.warn(
        #                         'Possible set difference at position %d' % (
        #                             source.tell() - 2),
        #                         FutureWarning, stacklevel=nested + 6
        #                     )
        #                 code2 = LITERAL, _ord(that)
        #             if code1[0] != LITERAL or code2[0] != LITERAL:
        #                 msg = "bad character range %s-%s" % (this, that)
        #                 raise source.error(msg, len(this) + 1 + len(that))
        #             lo = code1[1]
        #             hi = code2[1]
        #             if hi < lo:
        #                 msg = "bad character range %s-%s" % (this, that)
        #                 raise source.error(msg, len(this) + 1 + len(that))
        #             setappend((RANGE, (lo, hi)))

        elif item_type is RANGE and isinstance(item_value, tuple) and len(item_value) == 2:
            assert _parent_type is IN, _parent_type
            lo, hi = item_value
            return '{}-{}'.format(re.escape(_chr(lo)), re.escape(_chr(hi)))

        # # Reference:
        #         else:
        #             if code1[0] is IN:
        #                 code1 = code1[1][0]
        #             setappend(code1)

        # # Reference:
        #     set = _uniq(set)
        #     # XXX: <fl> should move set optimization to compiler!
        #     if _len(set) == 1 and set[0][0] is LITERAL:
        #         # optimization
        #         if negate:
        #             subpatternappend((NOT_LITERAL, set[0][1]))
        elif item_type is NOT_LITERAL:
            raise Exception("TODO", dict(case='NOT_LITERAL', value=rast))

            # # Reference:
            #     else:
            #         subpatternappend(set[0])
            # else:
            #     if negate:
            #         set.insert(0, (NEGATE, None))
        elif item_type is NEGATE and item_value is None:
            assert _parent_type is IN, _parent_type
            return '^'
            # # Reference:
            #     # charmap optimization can't be added here because
            #     # global flags still are not known
            #     subpatternappend((IN, set))
        elif item_type is IN:
            # XXXXX: needs re-checking
            return '[{}]'.format(''.join(
                rast_to_pattern(child, **kwargs)
                for child in item_value))

        # # Reference:
        # elif this in REPEAT_CHARS:
        #     # repeat previous item
        #     here = source.tell()
        #     if this == "?":
        #         min, max = 0, 1
        #     elif this == "*":
        #         min, max = 0, MAXREPEAT
        #     elif this == "+":
        #         min, max = 1, MAXREPEAT
        #     elif this == "{":
        #         if source.next == "}":
        #             subpatternappend((LITERAL, _ord(this)))
        #             continue
        #         min, max = 0, MAXREPEAT
        #         lo = hi = ""
        #         while source.next in DIGITS:
        #             lo += sourceget()
        #         if sourcematch(","):
        #             while source.next in DIGITS:
        #                 hi += sourceget()
        #         else:
        #             hi = lo
        #         if not sourcematch("}"):
        #             subpatternappend((LITERAL, _ord(this)))
        #             source.seek(here)
        #             continue
        #         if lo:
        #             min = int(lo)
        #             if min >= MAXREPEAT:
        #                 raise OverflowError("the repetition number is too large")
        #         if hi:
        #             max = int(hi)
        #             if max >= MAXREPEAT:
        #                 raise OverflowError("the repetition number is too large")
        #             if max < min:
        #                 raise source.error("min repeat greater than max repeat",
        #                                    source.tell() - here)
        #     else:
        #         raise AssertionError("unsupported quantifier %r" % (char,))
        #     # figure out which item to repeat
        #     if subpattern:
        #         item = subpattern[-1:]
        #     else:
        #         item = None
        #     if not item or item[0][0] is AT:
        #         raise source.error("nothing to repeat",
        #                            source.tell() - here + len(this))
        #     if item[0][0] in _REPEATCODES:
        #         raise source.error("multiple repeat",
        #                            source.tell() - here + len(this))
        #     if item[0][0] is SUBPATTERN:
        #         group, add_flags, del_flags, p = item[0][1]
        #         if group is None and not add_flags and not del_flags:
        #             item = p
        #     if sourcematch("?"):
        #         subpattern[-1] = (MIN_REPEAT, (min, max, item))
        # # Reference:
        #     else:
        #         subpattern[-1] = (MAX_REPEAT, (min, max, item))
        elif item_type is MAX_REPEAT or item_type is MIN_REPEAT and len(item_value) == 3:
            min_repeat, max_repeat, child = item_value
            if min_repeat == 0 and max_repeat == 1:
                modifier = '?'
            elif min_repeat == 0 and max_repeat is MAXREPEAT:
                modifier = '*'
            elif min_repeat == 1 and max_repeat is MAXREPEAT:
                modifier = '+'
            elif min_repeat == max_repeat:  # `{a}`
                modifier = '{{{}}}'.format(min_repeat)
            else:  # `{a,b}`
                modifier = '{{{},{}}}'.format(
                    '' if min_repeat == 0 else min_repeat,
                    '' if max_repeat is MAXREPEAT else max_repeat)
            return '{}{}{}'.format(
                _ensure_grouped(rast_to_pattern(child, **kwargs), source=child),
                modifier,
                # e.g. `.{4,}?` i.e. `(?:|.{4,})`
                '?' if item_type is MIN_REPEAT else '')

        # # Reference:
        # elif this == ".":
        #     subpatternappend((ANY, None))
        elif item_type is ANY and item_value is None:
            return '.'

        # # Reference:
        # elif this == "(":

            # # Reference:
            # start = source.tell() - 1
            # group = True
            # name = None
            # add_flags = 0
            # del_flags = 0
            # if sourcematch("?"):
            #     # options
            #     char = sourceget()
            #     if char is None:
            #         raise source.error("unexpected end of pattern")
            #     if char == "P":
            #         # python extensions
            #         if sourcematch("<"):
            #             # named group: skip forward to end of name
            #             name = source.getuntil(">", "group name")
            #             if not name.isidentifier():
            #                 msg = "bad character in group name %r" % name
            #                 raise source.error(msg, len(name) + 1)
            #         elif sourcematch("="):
            #             # named backreference
            #             name = source.getuntil(")", "group name")
            #             if not name.isidentifier():
            #                 msg = "bad character in group name %r" % name
            #                 raise source.error(msg, len(name) + 1)
            #             gid = state.groupdict.get(name)
            #             if gid is None:
            #                 msg = "unknown group name %r" % name
            #                 raise source.error(msg, len(name) + 1)
            #             if not state.checkgroup(gid):
            #                 raise source.error("cannot refer to an open group",
            #                                    len(name) + 1)
            #             state.checklookbehindgroup(gid, source)
            #             subpatternappend((GROUPREF, gid))
            #             continue

            # # Reference:
            #         else:
            #             char = sourceget()
            #             if char is None:
            #                 raise source.error("unexpected end of pattern")
            #             raise source.error("unknown extension ?P" + char,
            #                                len(char) + 2)
            #     elif char == ":":
            #         # non-capturing group
            #         group = None
            #     elif char == "#":
            #         # comment
            #         while True:
            #             if source.next is None:
            #                 raise source.error("missing ), unterminated comment",
            #                                    source.tell() - start)
            #             if sourceget() == ")":
            #                 break
            #         continue

            # # Reference:
            #     elif char in "=!<":
            #         # lookahead assertions
            #         dir = 1
            #         if char == "<":
            #             char = sourceget()
            #             if char is None:
            #                 raise source.error("unexpected end of pattern")
            #             if char not in "=!":
            #                 raise source.error("unknown extension ?<" + char,
            #                                    len(char) + 2)
            #             dir = -1 # lookbehind
            #             lookbehindgroups = state.lookbehindgroups
            #             if lookbehindgroups is None:
            #                 state.lookbehindgroups = state.groups
            #         p = _parse_sub(source, state, verbose, nested + 1)
            #         if dir < 0:
            #             if lookbehindgroups is None:
            #                 state.lookbehindgroups = None
            #         if not sourcematch(")"):
            #             raise source.error("missing ), unterminated subpattern",
            #                                source.tell() - start)
            #         if char == "=":
            #             subpatternappend((ASSERT, (dir, p)))
            #         else:
            #             subpatternappend((ASSERT_NOT, (dir, p)))
            #         continue

            # # Reference:
            #     elif char == "(":
            #         # conditional backreference group
            #         condname = source.getuntil(")", "group name")
            #         if condname.isidentifier():
            #             condgroup = state.groupdict.get(condname)
            #             if condgroup is None:
            #                 msg = "unknown group name %r" % condname
            #                 raise source.error(msg, len(condname) + 1)
            #         else:
            #             try:
            #                 condgroup = int(condname)
            #                 if condgroup < 0:
            #                     raise ValueError
            #             except ValueError:
            #                 msg = "bad character in group name %r" % condname
            #                 raise source.error(msg, len(condname) + 1) from None
            #             if not condgroup:
            #                 raise source.error("bad group number",
            #                                    len(condname) + 1)
            #             if condgroup >= MAXGROUPS:
            #                 msg = "invalid group reference %d" % condgroup
            #                 raise source.error(msg, len(condname) + 1)
            #         state.checklookbehindgroup(condgroup, source)
            #         item_yes = _parse(source, state, verbose, nested + 1)
            #         if source.match("|"):
            #             item_no = _parse(source, state, verbose, nested + 1)
            #             if source.next == "|":
            #                 raise source.error("conditional backref with more than two branches")
            #         else:
            #             item_no = None
            #         if not source.match(")"):
            #             raise source.error("missing ), unterminated subpattern",
            #                                source.tell() - start)
            #         subpatternappend((GROUPREF_EXISTS, (condgroup, item_yes, item_no)))
            #         continue

            # # Reference:
            #     elif char in FLAGS or char == "-":
            #         # flags
            #         flags = _parse_flags(source, state, char)
            #         if flags is None:  # global flags
            #             if not first or subpattern:
            #                 import warnings
            #                 warnings.warn(
            #                     'Flags not at the start of the expression %r%s' % (
            #                         source.string[:20],  # truncate long regexes
            #                         ' (truncated)' if len(source.string) > 20 else '',
            #                     ),
            #                     DeprecationWarning, stacklevel=nested + 6
            #                 )
            #             if (state.flags & SRE_FLAG_VERBOSE) and not verbose:
            #                 raise Verbose
            #             continue

            # # Reference:
            #         add_flags, del_flags = flags
            #         group = None
            #     else:
            #         raise source.error("unknown extension ?" + char,
            #                            len(char) + 1)

            # # Reference:
            # # parse group contents
            # if group is not None:
            #     try:
            #         group = state.opengroup(name)
            #     except error as err:
            #         raise source.error(err.msg, len(name) + 1) from None
            # sub_verbose = ((verbose or (add_flags & SRE_FLAG_VERBOSE)) and
            #                not (del_flags & SRE_FLAG_VERBOSE))
            # p = _parse_sub(source, state, sub_verbose, nested + 1)

            # # Reference glue: `_parse_sub` source:
            # def _parse_sub(source, state, verbose, nested):
            # # parse an alternation: a|b|c

            # # Reference: `_parse_sub` source:
            # items = []
            # itemsappend = items.append
            # sourcematch = source.match
            # start = source.tell()
            # while True:
            #     itemsappend(_parse(source, state, verbose, nested + 1,
            #                        not nested and not items))
            #     if not sourcematch("|"):
            #         break

            # # Reference: `_parse_sub` source:
            # if len(items) == 1:
            #     return items[0]

            # # Reference: `_parse_sub` source:
            # subpattern = SubPattern(state)

            # # Reference: `_parse_sub` source:
            # # check if all items share a common prefix
            # while True:
            #     prefix = None
            #     for item in items:
            #         if not item:
            #             break
            #         if prefix is None:
            #             prefix = item[0]
            #         elif item[0] != prefix:
            #             break
            #     else:
            #         # all subitems start with a common "prefix".
            #         # move it out of the branch
            #         for item in items:
            #             del item[0]
            #         subpattern.append(prefix)
            #         continue # check next one
            #     break

            # # Reference: `_parse_sub` source:
            # # check if the branch can be replaced by a character set
            # set = []
            # for item in items:
            #     if len(item) != 1:
            #         break
            #     op, av = item[0]
            #     if op is LITERAL:
            #         set.append((op, av))
            #     elif op is IN and av[0][0] is not NEGATE:
            #         set.extend(av)
            #     else:
            #         break
            # else:
            #     # we can store this as a character set instead of a
            #     # branch (the compiler may optimize this even more)
            #     subpattern.append((IN, _uniq(set)))
            #     return subpattern

            # # Reference: `_parse_sub` source:
            # subpattern.append((BRANCH, (None, items)))
            # return subpattern

        elif item_type is BRANCH and isinstance(item_value, tuple) and len(item_value) == 2 and item_value[0] is None:
            _, children = item_value
            # XXXXXXX: needs re-checking:
            result = '|'.join(
                rast_to_pattern(child, **kwargs)
                for child in children)
            # Tricky: branch inside subpattern does not require extra parentheses.
            if _parent_type is not SUBPATTERN:
                result = '(?:{})'.format(result)
            return result

            # # Reference:
            # if not source.match(")"):
            #     raise source.error("missing ), unterminated subpattern",
            #                        source.tell() - start)
            # if group is not None:
            #     state.closegroup(group, p)
            # subpatternappend((SUBPATTERN, (group, add_flags, del_flags, p)))
        elif item_type is SUBPATTERN and isinstance(item_value, tuple) and len(item_value) == 4:
            group, add_flags, del_flags, child = item_value
            flags = _flags_to_list(add_flags)
            if del_flags:
                flags.append('-')
                flags.extend(_flags_to_list(del_flags))

            if group is None:
                flags.append(':')
            else:
                group_to_name = kwargs.get('group_to_name')
                name = group_to_name.get(group)
                if name:
                    # NOTE: making a named group with changed flags is syntaxically impossible.
                    assert not flags
                    flags = ['P<{}>'.format(name)] + flags

            return '({}{})'.format(
                '?{}'.format(''.join(flags)) if flags else '',
                rast_to_pattern(child, **kwargs))

        # # Reference:
        # elif this == "^":
        #     subpatternappend((AT, AT_BEGINNING))
        elif item_type is AT and item_value is AT_BEGINNING:
            return '^'

        # # Reference:
        # elif this == "$":
        #     subpatternappend((AT, AT_END))
        elif item_type is AT and item_value is AT_END:
            return '$'

        # # Reference:
        # else:
        #     raise AssertionError("unsupported special character %r" % (char,))

    # # Reference:
    # # unpack non-capturing groups
    # for i in range(len(subpattern))[::-1]:
    #     op, av = subpattern[i]
    #     if op is SUBPATTERN:
    #         group, add_flags, del_flags, p = av
    #         if group is None and not add_flags and not del_flags:
    #             subpattern[i: i+1] = p

    # # Reference:
    # return subpattern
    raise Exception("TODO", dict(case="unknown", cls=rast.__class__, value=rast))


def cutoff_rast(rast, **kwargs):
    """
    Yield all smaller expressions for the specified parsed regular expressions.

    Intended for figuring out where the matching fails.
    """

    def make_cutoff_obj(lst):
        cutoff_obj = copy.copy(rast)
        cutoff_obj.pattern = copy.copy(cutoff_obj.pattern)
        # TODO?: mockup the cutoff_obj.pattern.str using rast_to_pattern
        cutoff_obj.pattern.str = None

        cutoff_obj.data = lst
        return cutoff_obj

    for idx in reversed(range(len(rast))):
        # Will start with the entire list and end with one item.
        cutoff = rast[:idx + 1]

        yield make_cutoff_obj(cutoff)

        last_child = cutoff[-1]
        if not last_child:
            continue
        # Recurse over the trees:
        if isinstance(last_child, tuple) and len(last_child) == 2:
            item_type, item_value = last_child

            # '...(abcd)' -> ['...(abcd)', '...(abc)', ..., '...(a)', '...']
            if item_type is SUBPATTERN and len(item_value) == 4:
                base_rast = list(cutoff)[:-1]

                subsubpattern_params = item_value[:-1]
                subsubpattern = item_value[-1]
                for subcutoff in cutoff_rast(subsubpattern, **kwargs):
                    cutoff_rec = base_rast + [(item_type, subsubpattern_params + (subcutoff,))]
                    yield make_cutoff_obj(cutoff_rec)

            # '...(abc|de)' -> ['...(abc|de)', '...(abc|d)', '...(ab|d)', '...(a|d)', '...']
            elif item_type is BRANCH and len(item_value) == 2 and item_value[0] is None:
                # item_value: `(None, [subpattern, subpattern, ...])`
                base_rast = list(cutoff)[:-1]

                subsubpattern_params = item_value[:-1]
                subsubpatterns = item_value[-1]
                subsubpatterns = subsubpatterns[:]
                for spidx, subsubpattern in reversed(list(enumerate(subsubpatterns))):
                    for subcutoff in cutoff_rast(subsubpattern, **kwargs):
                        # Mutating so that when a subsubpattern is exhausted it
                        # remains in the minimal form.
                        subsubpatterns[spidx] = subcutoff
                        cutoff_rec = base_rast + [(item_type, subsubpattern_params + (subsubpatterns[:],))]
                        yield make_cutoff_obj(cutoff_rec)


def normalize_pattern(pattern, flags=0):
    rast = sre_parse.parse(pattern, flags=flags)
    return rast_to_pattern(rast)


def find_matching_subregexes(pattern, string, flags=0, verbose=False):
    rast = sre_parse.parse(pattern, flags=flags)
    for cutoff in cutoff_rast(rast):
        subpattern = sre_compile.compile(cutoff, flags=flags)
        match = subpattern.search(string)
        if match:
            # Otherwise unusable:
            pattern_round = re.compile(rast_to_pattern(cutoff), flags=flags)
            if verbose:
                yield dict(
                    cutoff=cutoff,
                    subpattern=subpattern,
                    pattern_round=pattern_round,
                    match=match,
                    substring=match.group(0),
                    substring_unmatched=string[match.end(0):],
                )
            else:
                yield pattern_round


def main():
    import sys
    rex = sys.argv[1]
    string = sys.argv[2]
    print('Regex: {!r}, string: {!r}'.format(rex, string), file=sys.stderr)
    for result in find_matching_subregexes(rex, string, verbose=True):
        print('{pattern_round.pattern!r} -> {substring!r}'.format(**result))


if __name__ == '__main__':
    main()



# coding: utf8

import time
import logging
from logging import handlers
from .base import to_bytes


__all__ = (
    'TaggedSysLogHandlerBase',
    'TaggedSysLogHandler',
)


class TaggedSysLogHandlerBase(handlers.SysLogHandler):
    """
    A version of SysLogHandler that adds a tag to the logged line
    so that it can be sorted by the syslog daemon into files.

    Generally equivalent to using a formatter but semantically more
    similar to FileHandler's `filename` parameter.

    Example rsyslog config:
    In `/etc/rsyslog.d/01-dynamic-app-logging.conf`:

        # WARN: global settings.
        $MaxMessageSize 2049k
        $EscapeControlCharactersOnReceive off  # In case your logs aren't quite text.
        $RepeatedMsgReduction off
        $SystemLogRateLimitInterval 0
        $SystemLogRateLimitBurst 0
        # File permissions:
        $umask 0000
        $FileOwner syslog
        $DirOwner syslog
        # Makes it possible to set a group on the logging directory to give
        # write access to it.
        $FileCreateMode 0664
        $DirCreateMode 2775
        $CreateDirs on

        $template LogFormatPlain,"%msg:2:$%\n"
        # Make it possible to specify the target filename in the syslog tag.
        # NOTE: The root directory creation has to be done outside syslog, as
        # syslog normally doesn't have access to `mkdir /var/log/something`.
        # NOTE: the '.log' is appended automatically.
        $template LogDynFileAUTO,"/var/log/%syslogtag:R,ERE,1,DFLT:file__([a-zA-Z0-9_/.-]*)--end%.log"
        :syslogtag, regex, "file__[a-zA-Z0-9_/.-]*" ?LogDynFileAUTO;LogFormatPlain

        # NOTE: ampersand-tilde tells rsyslog to drop the lines that passed the
        # last filter line; i.e. kind-of 'propagate=False'
        & ~

    """

    def __init__(self, *args, **kwargs):
        syslog_tag = kwargs.pop('syslog_tag')
        syslog_tag = to_bytes(syslog_tag)
        self.syslog_tag = syslog_tag
        super(TaggedSysLogHandler, self).__init__(*args, **kwargs)

    def format(self, *args, **kwargs):
        res = super(TaggedSysLogHandler, self).format(*args, **kwargs)
        assert isinstance(res, bytes)
        return self.syslog_tag + " " + res


class TaggedSysLogHandler(TaggedSysLogHandlerBase):
    """
    An addition to TaggedSysLogHandlerBase that sets the SO_SNDBUF to a large
    value to allow large log lines.
    """

    _sndbuf_size = 5 * 2**20  # 5 MiB

    def __init__(self, *args, **kwargs):
        self._sbdbuf_size = kwargs.pop('sbdbuf_size', self._sndbuf_size)
        super(TaggedSysLogHandler, self).__init__(*args, **kwargs)
        self.configure_socket(self.socket)

    def configure_socket(self, sock):
        import socket
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, self._sndbuf_size)



# coding: utf8
"""
Requests for humans that have to deal with lots of stuff.
"""
# pylint: disable=arguments-differ,too-many-arguments

from __future__ import division, absolute_import, print_function, unicode_literals

import sys
import logging
import functools

import requests
import requests.adapters
import requests.exceptions
from requests.packages.urllib3.util import Retry  # pylint: disable=import-error
from pyaux.minisix import text_type, urllib_parse
from pyaux.base import (
    find_caller,
    split_dict, dict_maybe_items,
    LazyRepr, ReprObj, to_bytes,
)

# Singleton session for connection pooling.
# WARNING: this involves caching; do `SESSION.close()` in a `finally:` block to clean it up.
SESSION = requests.Session()


UNDEF = ReprObj('Unspecified')
AUTO = ReprObj('Autoselected')


def _is_special(value, none=True, undef=True, auto=True):
    """
    Check whether `value` is one of the special values used in this module.
    """
    if none and value is None:
        return True
    if undef and value is UNDEF:
        return True
    if auto and value is AUTO:
        return True
    return False


def configure_session(
        session,
        retries_total=5,
        retries_backoff_factor=0.5,
        retries_status_forcelist=(500, 502, 503, 504, 521),
        retries_method_whitelist=frozenset((
            'HEAD', 'OPTIONS', 'TRACE', 'GET',
            'POST', 'PUT', 'PATCH', 'DELETE')),
        pool_connections=30,
        pool_maxsize=30,
        raise_on_status=False,
        **kwargs):
    """
    Configure session with significant retries.

    WARNING: by default, retries mutating requests too. Use only when
    idempotence is expected (or at-least-once delivery is generally okay).
    """
    retry_conf = Retry(
        total=retries_total,
        backoff_factor=retries_backoff_factor,
        status_forcelist=retries_status_forcelist,
        method_whitelist=retries_method_whitelist,
        # https://stackoverflow.com/a/43496895
        raise_on_status=raise_on_status,
    )

    for prefix in ('http://', 'https://'):
        session.mount(
            prefix,
            requests.adapters.HTTPAdapter(
                max_retries=retry_conf,
                pool_connections=pool_connections,
                pool_maxsize=pool_maxsize,
                **kwargs))

    return session


def _cut(data, length, marker):
    if not length:
        return data
    actual_length = length - len(marker)
    assert actual_length >= 2, "can't cut to/below marker length"
    if len(data) < length:
        return data
    right_length = actual_length // 2
    left_length = actual_length - right_length
    return data[:left_length] + marker + data[right_length:]


SESSION_ZEALOUS = configure_session(requests.Session())


class RequesterBase(object):

    apply_environment = True
    length_cut_marker = b'...'
    _prepare_request_keys = frozenset((
        'method', 'url',
        'headers', 'auth', 'cookies',
        'params',
        'data', 'files', 'json',
        'hooks',
    ))
    send_request_keys = frozenset((
        'proxies', 'cert', 'verify',
        'stream', 'timeout',
        'allow_redirects',
    ))

    # Warning: argument defaults are also duplicated in `Requester.__init__`.
    def __init__(self, session=True):
        if session is True:
            session = SESSION
        elif _is_special(session):
            session = requests.Session()
        self.session = session
        self.logger = logging.getLogger('req')

    # Warning: kwargs defaults are also duplicated in `Requester.request`.
    @staticmethod
    def _preprepare_parameters(kwargs):
        """
        Separated out for easier overriding of `prepare_parameters` by subclasses.
        """
        headers = kwargs.get('headers')
        # For common convenience, commonize the headers:
        if _is_special(headers):
            headers = {}
        else:
            # TODO?: OrderedDict?
            headers = {
                key.lower().replace('_', '-'): value
                for key, value in dict_maybe_items(headers)}
        kwargs['headers'] = headers

        return kwargs

    # Warning: kwargs defaults are also duplicated in `Requester.request`.
    def _prepare_parameters(self, kwargs):  # pylint: disable=no-self-use
        return kwargs

    def request(self, url, **kwargs):
        kwargs['url'] = url
        request, send_kwargs = self._prepare_request(kwargs)
        return self.send_request(request, **send_kwargs)

    # Warning: kwargs defaults are also duplicated in `Requester.request`.
    def _prepare_request(self, kwargs, run_processing=True):
        """
        ...

        Not to be confused with `Session.prepare_request`, which takes a
        request and returns only a request.
        """
        session = kwargs.pop('session', None) or self.session

        if run_processing:
            kwargs = self._preprepare_parameters(kwargs)
            kwargs = self._prepare_parameters(kwargs)

        assert kwargs.get('method')
        assert kwargs.get('url')

        prepare_keys = self._prepare_request_keys
        prepare_kwargs, send_kwargs = split_dict(
            kwargs, lambda key, value: key in prepare_keys)

        send_keys = self.send_request_keys
        send_kwargs, unknown_kwargs = split_dict(
            send_kwargs, lambda key, value: key in send_keys)

        if unknown_kwargs:
            raise ValueError("Unknown request arguments", unknown_kwargs)

        request = requests.Request(**prepare_kwargs)
        request = session.prepare_request(request)
        if self.apply_environment:
            send_kwargs_overrides = send_kwargs
            # http://docs.python-requests.org/en/master/user/advanced/#prepared-requests
            send_kwargs = session.merge_environment_settings(
                request.url, {}, None, None, None)
            send_kwargs.update(send_kwargs_overrides)
        send_kwargs['session'] = session
        return request, send_kwargs

    def send_request(self, request, **kwargs):
        """
        ...

        Essentially a wrap of `self.session.send_request` but the session
        object can be specified in arguments.
        """
        session = kwargs.pop('session', None) or self.session
        return session.send(request, **kwargs)

    def __call__(self, *args, **kwargs):
        """ See `request` method """
        kwargs['call_extra_depth'] = (kwargs.get('call_extra_depth') or 0) + 1
        return self.request(*args, **kwargs)


class RequesterSessionWrap(RequesterBase):

    def __getattr__(self, key):
        return getattr(self.session, key)


class RequesterVerbMethods(RequesterBase):
    """
    Add the http-method-named methods to the class, wrapping `self.request`.
    """

    def get(self, *args, **kwargs):
        """ See `request` method """
        kwargs['call_extra_depth'] = (kwargs.get('call_extra_depth') or 0) + 1
        kwargs['method'] = 'get'
        return self.request(*args, **kwargs)

    def options(self, *args, **kwargs):
        """ See `request` method """
        kwargs['call_extra_depth'] = (kwargs.get('call_extra_depth') or 0) + 1
        kwargs['method'] = 'options'
        return self.request(*args, **kwargs)

    def head(self, *args, **kwargs):
        """ See `request` method """
        kwargs['call_extra_depth'] = (kwargs.get('call_extra_depth') or 0) + 1
        kwargs['method'] = 'head'
        return self.request(*args, **kwargs)

    def post(self, *args, **kwargs):
        """ See `request` method """
        kwargs['call_extra_depth'] = (kwargs.get('call_extra_depth') or 0) + 1
        kwargs['method'] = 'post'
        return self.request(*args, **kwargs)

    def put(self, *args, **kwargs):
        """ See `request` method """
        kwargs['call_extra_depth'] = (kwargs.get('call_extra_depth') or 0) + 1
        kwargs['method'] = 'put'
        return self.request(*args, **kwargs)

    def patch(self, *args, **kwargs):
        """ See `request` method """
        kwargs['call_extra_depth'] = (kwargs.get('call_extra_depth') or 0) + 1
        kwargs['method'] = 'patch'
        return self.request(*args, **kwargs)

    def delete(self, *args, **kwargs):
        """ See `request` method """
        kwargs['call_extra_depth'] = (kwargs.get('call_extra_depth') or 0) + 1
        kwargs['method'] = 'delete'
        return self.request(*args, **kwargs)


class RequesterDefaults(RequesterBase):

    # Warning: argument defaults are also duplicated in `Requester.__init__`.
    def __init__(
            self, default_timeout=5.0, default_write_method='post',
            default_allow_redirects=AUTO,
            **kwargs):
        self.default_timeout = default_timeout
        self.default_write_method = default_write_method
        self.default_allow_redirects = default_allow_redirects
        super(RequesterDefaults, self).__init__(**kwargs)

    # Warning: kwargs defaults are also duplicated in `Requester.request`.
    def _prepare_parameters(self, kwargs):
        # NOTE: will leave explicit `timeout=None` as-is.
        if _is_special(kwargs.get('timeout'), none=False):
            kwargs.update(timeout=self.default_timeout)

        # Determine the method by the presence of `data` / `files`.
        if _is_special(kwargs.get('method')):
            if kwargs.get('data') is not None or kwargs.get('files') is not None:
                kwargs['method'] = self.default_write_method
            else:
                kwargs['method'] = 'get'

        if _is_special(kwargs.get('allow_redirects')):
            if self.default_allow_redirects == 'auto' or self.default_allow_redirects is AUTO:
                # From `requests.get` logic.
                kwargs['allow_redirects'] = kwargs['method'] in ('get', 'options')
            else:
                kwargs['allow_redirects'] = self.default_allow_redirects

        return super(RequesterDefaults, self)._prepare_parameters(kwargs)


class RequesterBaseUrl(RequesterBase):
    """ Common relative URL support for RequesterBase """

    # Warning: argument defaults are also duplicated in `Requester.__init__`.
    def __init__(self, base_url=None, **kwargs):
        self.base_url = base_url
        super(RequesterBaseUrl, self).__init__(**kwargs)

    # Warning: kwargs defaults are also duplicated in `Requester.request`.
    def _prepare_parameters(self, kwargs):
        if self.base_url:
            kwargs['url'] = urllib_parse.urljoin(
                self.base_url,
                kwargs['url'])  # required parameter
        return super(RequesterBaseUrl, self)._prepare_parameters(kwargs)


class RequesterContentType(RequesterBase):
    """
    `requests` itself allows a `json=...` parameter to the methods.

    However, that is a bad design decision.

    This mixin adds a support for 'content_type' paremeter which defines how
    the body should be serialized.

    WARNING: it also sets the default content type to `'json'`.
    """

    # Warning: argument defaults are also duplicated in `Requester.__init__`.
    def __init__(self, default_content_type='json', **kwargs):
        self.default_content_type = default_content_type
        super(RequesterContentType, self).__init__(**kwargs)

    # Warning: kwargs defaults are also duplicated in `Requester.request`.
    def _prepare_parameters(self, kwargs):
        content_type = kwargs.pop('content_type', None)
        if _is_special(content_type):
            content_type = self.default_content_type
        data = kwargs.get('data')
        json_data = kwargs.pop('json', None)
        if json_data is not None:
            raise Exception((
                "Please don't use `json=...` keyword here."
                " Use `content_type='json'` (likely, already the default)."))
        if data is not None and content_type is not None:
            # # Maybe:
            # if content_type == 'multipart':
            #     kwargs.pop('data', None)
            #     kwargs['files'] = data
            data, content_type_header = self.serialize_data(
                data, content_type=content_type, context=kwargs)
            if data is not None:
                kwargs['data'] = data
                if content_type_header is not None:
                    kwargs['headers']['content-type'] = content_type_header
        return super(RequesterContentType, self)._prepare_parameters(kwargs)

    json_content_type = 'application/json; charset=utf-8'

    def serialize_data(self, data, content_type, context=None):
        """
        Data object `data`, desired serialization `content_type`
        ->
        serialized_data, content_type_header

        NOTE: `content_type=None` has a special meaning 'pass the data as-is'.
        """
        if _is_special(content_type):
            return data, None

        # TODO: check `data` for being a file-like or a bytestream or such.

        # # Maybe:
        # if content_type == 'urlencode':
        #     return None, data
        if content_type == 'json':
            return self._serialize_data_json(data, context=context)
        if content_type == 'ujson':  # Special case for dangerous performance.
            return self._serialize_data_ujson(data, context=context)
        raise Exception("Unknown data serialization content_type", content_type)

    def _serialize_data_json(self, data, context=None):  # pylint: disable=unused-argument
        """ Overridable point for json-serialization customization. """
        from .anyjson import json_dumps
        data = to_bytes(json_dumps(data))
        return data, self.json_content_type

    def _serialize_data_ujson(self, data, context=None):  # pylint: disable=unused-argument
        import ujson
        # pylint: disable=c-extension-no-member
        data = to_bytes(ujson.dumps(data, ensure_ascii=False))
        return data, self.json_content_type


class RequesterMeta(RequesterBase):

    # Warning: argument defaults are also duplicated in `Requester.__init__`.
    def __init__(self, collect_call_info=True, call_info_in_ua=True, **kwargs):
        self.collect_call_info = collect_call_info
        self.call_info_in_ua = call_info_in_ua
        super(RequesterMeta, self).__init__(**kwargs)

    # Warning: kwargs defaults are also duplicated in `Requester.request`.
    def _prepare_parameters(self, kwargs):
        call_info = kwargs.pop('call_info', None)
        if call_info and self.call_info_in_ua:
            # `call_info`: `(cfile, cline, cfunc)`.
            kwargs['headers']['user-agent'] = '{}, {}:{}: {}'.format(
                kwargs['headers'].get('user-agent') or '',
                *call_info)
        return super(RequesterMeta, self)._prepare_parameters(kwargs)

    def request(self, url, **kwargs):
        """
        ...

        WARNING: should either be the entry-point method or be wrapped with

            kwargs['call_extra_depth'] = (kwargs.get('call_extra_depth') or 0) + 1
        """
        call_extra_depth = kwargs.pop('call_extra_depth', 0)
        if self.collect_call_info and kwargs.get('call_info') is None:
            # Could actually filter out the current module, but that's costly and palliative.
            kwargs['call_info'] = find_caller(extra_depth=call_extra_depth + 1)
        return super(RequesterMeta, self).request(url, **kwargs)


class RequesterAutoRaiseForStatus(RequesterBase):

    raise_with_content = True
    raise_content_cap = 1800  # in bytes.
    response_exception_cls = requests.exceptions.HTTPError

    # Warning: kwargs defaults are also duplicated in `Requester.request`.
    def _prepare_request(self, kwargs, **etcetera):
        # This could be done by auto-merging `send_request_keys` by the MRO /
        # property, but for a single case this is easier:
        require = kwargs.pop('require', True)
        request, send_kwargs = (
            super(RequesterAutoRaiseForStatus, self)
            ._prepare_request(kwargs, **etcetera))
        send_kwargs['require'] = require
        return request, send_kwargs

    def send_request(self, request, **kwargs):
        require = kwargs.pop('require', True)
        response = super(RequesterAutoRaiseForStatus, self).send_request(request, **kwargs)
        if require:
            self.raise_for_status(response)
        return response

    def raise_for_status(self, response):
        """ `response.raise_for_status()` or similar """
        if not self.raise_with_content:
            response.raise_for_status()
            return
        if response.ok:
            return
        self.raise_for_status_forced(response)

    def _get_raise_content(self, response):
        # TODO?: support streaming?
        # Note: cutting the bytestream.
        content = response.content
        content = _cut(
            content,
            length=self.raise_content_cap,
            marker=self.length_cut_marker)
        return content

    def raise_for_status_forced(self, response):
        """
        `response.raise_for_status()` but for any status.

        Also attempts to put a piece of the response body into the error.
        """
        try:
            content = self._get_raise_content(response)
        except Exception as exc:
            self.logger.warning("_get_raise_content failed: %r", exc)
            raise self.response_exception_cls(
                "Status Error: {} {}".format(
                    response.status_code, response.reason),
                response=response)
        raise self.response_exception_cls(
            "Status Error: {} {}: {}".format(
                response.status_code, response.reason, content),
            response=response)


class RequesterLog(RequesterBase):

    log_response_headers = True
    logging_url_cap = 1800

    def send_request(self, request, **kwargs):
        self.log_before(request, **kwargs)
        try:
            response = super(RequesterLog, self).send_request(request, **kwargs)
        except Exception as exc:
            exc_info = sys.exc_info()
            self.log_exc(request, exc, exc_info=exc_info, **kwargs)
            raise
        self.log_after(request, response, **kwargs)
        return response

    def request_for_logging(self, request):
        """ Make a log string out of a request """
        url = request.url
        url_cap = self.logging_url_cap
        if url_cap and len(url) > url_cap:
            url = '{}{}'.format(url[:url_cap], self.length_cut_marker)

        data_info = ''
        if request.body:
            data_info = "  data_len={}".format(len(request.body))

        return "{method} {url}{data_info}".format(
            method=request.method.upper(), url=url, data_info=data_info)

    def request_for_logging_lazy(self, request):
        """ ... exactly what it says """
        return LazyRepr(functools.partial(self.request_for_logging, request))

    def response_for_logging(self, response):
        """ Make a log string out of a response """
        pieces = [
            response.status_code,
            response.request.method,
            response.url,
            '    {}b'.format(len(response.content or '')),
        ]
        try:
            elapsed = "in {:.3f}s".format(response.elapsed.total_seconds())
        except Exception:  # pylint: disable=broad-except
            elapsed = "in ???s"
        pieces.append(elapsed)
        if self.log_response_headers:
            pieces.append('    response.headers={!r}'.format(response.headers))
        return ' '.join(
            text_type(piece) if not isinstance(piece, text_type) else piece
            for piece in pieces)

    # pylint: disable=unused-argument
    def log_before(self, request, level=logging.DEBUG, **kwargs):
        """ Hook for logging before a request is sent """
        if not self.logger.isEnabledFor(level):
            return
        self.logger.debug(
            "Sending request: %s",
            self.request_for_logging_lazy(request))

    # pylint: disable=unused-argument
    def log_after(self, request, response, level=logging.INFO, **kwargs):
        """ Hook for logging after a response is received """
        if not self.logger.isEnabledFor(level):
            return
        self.logger.log(
            level,
            "Response: %s    -> %s",
            self.request_for_logging_lazy(request),
            self.response_for_logging(response))

    # pylint: disable=unused-argument
    def log_exc(self, request, exc, exc_info=None, level=logging.ERROR, **kwargs):
        """ Log an exception that occured when getting a response """
        if not self.logger.isEnabledFor(level):
            return

        response_str = ''
        response = getattr(exc, 'response', None)
        if response is not None:
            response_str = ' ({})'.format(
                self.response_for_logging(response))

        self.logger.log(
            level,
            "Request exception: %s -> %r%s",
            self.request_for_logging_lazy(request),
            exc,
            response_str,
            exc_info=exc_info or True)


class Requester(
        RequesterLog, RequesterAutoRaiseForStatus, RequesterMeta,
        RequesterContentType, RequesterBaseUrl, RequesterDefaults,
        RequesterVerbMethods, RequesterSessionWrap, RequesterBase):

    def __init__(
            self,
            session=True,
            default_timeout=5.0, default_write_method='post', default_allow_redirects=AUTO,
            base_url=None,
            default_content_type='json',
            collect_call_info=True, call_info_in_ua=True,
            **kwargs):
        """
        ...

        WARNING: by default it uses a singleton session; if you want to put default
        headers on it, specify `session=None`.

        WARNING: by default, adds caller filename, line number, function name
        to the user-agent header.

        :param session: session object to use; additionally:
          `None`: make a new session;
          `True`: use a singleton session;

        :param default_timeout: ...

        :param default_write_method: what method to use when request body is specified.

        :param default_allow_redirects: whether return or follow a redirect;
          'auto' / `Autoselected` means "only for safe request methods".

        :param base_url: URL to use for relative links, in HTML /
          urllib.parse.urljoin semantics.

        :param default_content_type: serialization to use for non-bytestream `data`
          (request body) by default.

        :param collect_call_info: collect the caller (stack) information at all.

        :param call_info_in_ua: add the stack information to user agent.
        """
        super(Requester, self).__init__(
            session=session,
            default_timeout=default_timeout,
            default_write_method=default_write_method,
            default_allow_redirects=default_allow_redirects,
            base_url=base_url,
            default_content_type=default_content_type,
            collect_call_info=collect_call_info,
            call_info_in_ua=call_info_in_ua,
            **kwargs)

    # TODO: make it possible to supply defaults for everything in `__init__`.
    # pylint: disable=unused-argument,too-many-locals
    def request(
            self, url, params=None, data=None,
            # *,  # someday, someday.
            method=None, headers=None, auth=None, cookies=None,
            content_type=None, files=None,
            hooks=None, proxies=None, cert=None,
            verify=True, stream=False,
            timeout=UNDEF, allow_redirects=None,
            call_extra_depth=0, call_info=None,
            session=None,
            require=True,
            **kwargs):
        """
        Send a request, return a response.

        :param url: URL to send the request to;
          can be relative to `base_url`.

        :param params: query-string contents; see `requests.request`.

        :param data: a serializable python object, or bytestream, or file-like
        object to send in the request body; see `requests.request`.

        :param require: whether a non-okay response should raise (automatic raise_for_status).

        :param method: HTTP method to use;
          default: 'get' if no request body is specified,
          `default_write_method` otherwise.

        :param headers: additional HTTP headers dict.

        :param auth: Basic/Digest/Custom HTTP Auth;
          e.g. an `(username, password) tuple for sending to the server in basic auth;
          see `requests.request`;
        http://docs.python-requests.org/en/master/user/advanced/#custom-authentication

        :param cookies: dict or CookieJar; see `requests.request`.

        :param content_type: which type to serialize the `data` in, if necessary;
          default: see `default_content_type` in `__init__`.

        :param files: streams to send in a multipart body; see `requests.request`.

        :param hooks: see `requests.request`;
        http://docs.python-requests.org/en/master/user/advanced/#event-hooks

        :param proxies: protocol to proxy address; see `requests.request`;
        http://docs.python-requests.org/en/master/user/advanced/#proxies

        :param cert: SSL client certificate; see `requests.request`.

        :param verify: server certificate verification; see `requests.request`.

        :param stream: allow chunked reading of the response body; see `requests.request`.

        :param timeout: single request attempt timeout; see `requests.request`;
          default: see `default_timeout` in `__init__`;
          `None` means "no timeout".

        :param allow_redirects: whether HTTP 301 / HTTP 302 should be followed;
          see `requests.request`;
          default: see `default_allow_redirects` in `__init__`.

        :param call_extra_depth: for call_info collection, skip this many stack
          items. Increment this by one when wrapping the method.

        :param call_info: excplicit `(filename, lineno, func)` to use as caller
          info (instead of `call_extra_depth`).

        :param session: requests.Session object to use; defaults to `self.session`.
        """
        return super(Requester, self).request(
            url,
            session=session,
            method=method, headers=headers, auth=auth, cookies=cookies, params=params,
            data=data, content_type=content_type, files=files,
            hooks=hooks,
            proxies=proxies, cert=cert, verify=verify,
            stream=stream, timeout=timeout, allow_redirects=allow_redirects,
            call_extra_depth=call_extra_depth + 1, call_info=call_info,
            require=require,
            **kwargs)


class APIRequester(Requester):
    """
    Requester slightly tuned for backend-to-APIs interaction.

    Includes retries, large timeout, and considers only a very small list of
    statuses as okay.

    Warning: by default it uses a singleton session; if you want to put default
    headers on it, specify `session=None`.
    """

    def __init__(self, session=True, **kwargs):
        if session is True:
            session = SESSION_ZEALOUS
        elif _is_special(session):
            session = configure_session(requests.Session())
        kwargs.setdefault('default_timeout', 120)
        kwargs.setdefault('default_allow_redirects', False)
        super(APIRequester, self).__init__(session=session, **kwargs)

    # '204' is often returned for `DELETE` requests.
    okay_statuses = (200, 201, 204)

    def raise_for_status(self, response):
        if response.status_code not in self.okay_statuses:
            self.raise_for_status_forced(response)


def test():
    from pyaux.runlib import init_logging
    init_logging(level=1)
    reqr = Requester()
    resp = reqr.get('https://example.com?param1=a', params=dict(param2='b'))
    print(resp)
    assert resp.ok
    resp = reqr(
        'https://example.com?param2=a',
        params=dict(param1='b'),
        headers={'Content-Type': 'nope, delete this'},
        data=dict(param3=['c', 1]))
    print(resp)


if __name__ == '__main__':
    test()



#!/usr/bin/env python

from __future__ import print_function, unicode_literals, absolute_import, division

import os
from setuptools import setup, find_packages

version = '2.9.0'  # should be the same as pyaux.__version__

LONG_DESCRIPTION = """
Collection of helpers and useful things for Python

"""

try:
    LONG_DESCRIPTION = (
        LONG_DESCRIPTION +
        open(os.path.join(os.path.dirname(__file__), 'README.rst'), 'rb').read().decode('utf-8'))
except Exception as _exc:
    print("Pkg-description error:", _exc)


setup_kwargs = dict(
    name='pyaux',
    version=version,
    description='pyaux',  ## XX
    long_description=LONG_DESCRIPTION,
    # classifiers=[],
    # keywords='...,...',
    author='HoverHell',
    author_email='hoverhell@gmail.com',
    url='https://gitlab.com/hoverhell/pyaux',
    download_url='https://gitlab.com/hoverhell/pyaux/-/archive/{0}/pyaux-{0}.tar.bz2'.format(version),
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'pylzcat = pyaux.lzcat:_lzcat_main',
            'pyunlzma = pyaux.lzcat:_lzcat_main',
            # NOTE: cannot name it 'lzma.py' as it will get imported
            # and break py3's zipfile
            'pylzma = pyaux.lzmah:_lzma_main',
            # And naming it 'pylzma' is confusing too, so provide a
            # preferred alias.
            'pyenlzma = pyaux.lzmah:_lzma_main',
            'fjson_yaml = pyaux.bin.fjson_yaml:main',
            'fyaml_json = pyaux.bin.fyaml_json:main',
            'fjson.py = pyaux.bin.fjson:main',
            'fmsgp_json = pyaux.bin.fmsgp_json:main',
            'f_convert = pyaux.bin.f_convert:main',
        ],
    },
    install_requires=[],
    extras_require=dict(
        # Things that are useful to simply have around:
        recommended=[
            'ipython', 'ipdb', 'PyYAML',
            'atomicfile', 'cdecimal',
            # 'requests', 'pycurl',
            'simplejson',
        ],
        # All things that are known to be used in some part of this
        # library or another.
        known=[
            'django',  # in the psql helper
            'Twisted',  # bunch of twisted stuff here
            'Cython',  # at least one pyx module
            'pandas',  # here and there
            'Pygments',  # json / yaml coloring
            'pylzma',  # helpers for it
            'simplejson',  # optional, for speed
            # 'pp',  # so special I won't even include it here
            'line_profiler',
            # 'pyzmq',  # also too rare
        ],
        req=[
            'requests[ssl]',
            'six',
        ],
    ),
    dependency_links=[
        # 'https://github.com/sashka/atomicfile/tarball/master#egg=atomicfile',  # on pypi now
    ],
    # package_data={},
    # include_package_data=True,
    # zip_safe=False,
)


if __name__ == '__main__':
    setup(**setup_kwargs)



""" Various (un)necessary stuff for runscripts """

from __future__ import print_function, unicode_literals, absolute_import, division

import os
import sys
import random
import time
import warnings
import logging
import signal
import atexit
import traceback
from six.moves import xrange


__all__ = [
    'init_logging',
    'sigeventer',
]


# TODO: lazy-imports
try:
    from pyaux.twisted_aux import (
        make_manhole_telnet,
        make_manhole,
    )
except ImportError:
    pass


def _make_short_levelnames(shortnum=True):
    """ Return a dict (levelnum -> levelname) with short names for logging.
    `shortnum`: also shorten all 'Level #' names to 'L##'.
    """
    _names = dict([
        (logging.DEBUG, 'DBG'),
        (logging.INFO, 'INFO'),  # d'uh
        (logging.WARN, 'WARN'),
        (logging.ERROR, 'ERR'),
        (logging.CRITICAL, 'CRIT'),
    ])
    if shortnum:
        for i in xrange(1, 100):
            _names.setdefault(i, "L%02d" % (i,))
    return _names


# Current attempt: use bytes in py2, unicode in py3 (i.e. subvert unicode_literals just for these).
BASIC_LOG_FORMAT = str("%(asctime)s: %(levelname)-13s: %(name)s: %(message)s")
BASIC_LOG_FORMAT_TD = str("%(asctime)s(+%(time_diff)5.3fs): %(levelname)-13s: %(name)s: %(message)s")


def init_logging(*args, **kwargs):
    """ Simple shorthand for neat and customizable logging init """
    _td = kwargs.pop('_td', False)
    # Support for https://pypi.python.org/pypi/coloredlogs
    # Also, notable: https://pypi.python.org/pypi/verboselogs (but no special support here at the moment)
    _try_coloredlogs = kwargs.pop('_try_coloredlogs', False)

    if _try_coloredlogs:
        try:
            import coloredlogs
        except Exception:
            coloredlogs = None
            _try_coloredlogs = False

    colored = kwargs.pop('colored', True)
    if colored and not _try_coloredlogs:
        from . import use_colorer
        use_colorer()

    short_levelnames = kwargs.pop('short_levelnames', True)
    if short_levelnames:
        _names = _make_short_levelnames()
        for lvl, name in _names.items():
            logging.addLevelName(lvl, str(name))

    kwargs.setdefault('level', logging.DEBUG)

    logformat = BASIC_LOG_FORMAT
    if _td:
        logformat = BASIC_LOG_FORMAT_TD
    kwargs.setdefault('format', logformat)

    if _try_coloredlogs:
        kwargs['fmt'] = kwargs['format']
        coloredlogs.install(*args, **kwargs)
    else:
        logging.basicConfig(*args, **kwargs)

    if _td:
        # XX: do the same for all `logging.Logger.manager.loggerDict.values()`?
        from .logging_annotators import time_diff_annotator
        flt = time_diff_annotator()
        logging.root.addFilter(flt)
        for logger in logging.Logger.manager.loggerDict.values():
            if hasattr(logger, 'addFilter'):
                logger.addFilter(flt)


def argless_wrap(fn):
    """ Wrap function to re-try calling it if calling it with arguments
    failed """

    def argless_internal(*ar, **kwa):
        try:
            return fn(*ar, **kwa)
        except TypeError as e:
            try:
                return fn()
            except TypeError as e2:
                #raise e  # - traceback-inconvenient
                raise  # - error-inconvenient

    return argless_internal


# convenience wrappers:
def _sysexit_wrap(n=None, f=None):
    return sys.exit()


def _atexit_wrap(n=None, f=None):
    return atexit._run_exitfuncs()


class ListSigHandler(list):

    def __init__(self, try_argless, ignore_exc, verbose):
        self.try_argless = try_argless
        self.ignore_exc = ignore_exc
        self.verbose = verbose

    def __call__(self, n, f):
        for func in reversed(self):
            try:
                if self.verbose:
                    print("ListSigHandler: running %r" % (func,))
                if self.try_argless:
                    func = argless_wrap(func)
                func(n, f)
            except Exception as e:
                if self.ignore_exc:
                    if self.verbose:
                        traceback.print_exc()
                    else:
                        # Still print something
                        print("Exception ignored: %r" % (e,))
                else:
                    raise


def sigeventer(add_defaults=True, add_previous=True, do_sysexit=True,
               try_argless=True, ignore_exc=True, verbose=False):
    """
    Puts one list-based handler for SIGINT and SIGTERM that can be `append`ed to.

    NOTE: arguments are ignored if it was called previously.

    :param add_defaults: add the `atexit` handler.
    :param add_previous: add the previously-set handlers (NOTE: will
        mix sigterm/sigint handlers if different).
    :param try_argless: re-call handled function without parameters if
        they raise TypeError.
    :param do_sysexit: do `sys.exit()` at the end of handler.
    :param ignore_exc: ...

    Use `signal.getsignal(signal.SIGINT).append(some_func)` to add a handler.
    Handlers are called in reverse order (first in, last out).
    """
    # XXXX/TODO: a version that only sets itself on one of the signals
    # (optionally).
    # Check if already done something like this:
    curhandler_int = signal.getsignal(signal.SIGINT)
    curhandler_term = signal.getsignal(signal.SIGTERM)

    if isinstance(curhandler_int, list) and isinstance(curhandler_term, list):
        # probaby us already; just return
        assert curhandler_int is curhandler_term, \
            "unexpected: different list-based term/int handlers"
        return curhandler_term

    the_handler = ListSigHandler(try_argless=try_argless, ignore_exc=ignore_exc, verbose=verbose)
    signal.signal(signal.SIGINT, the_handler)
    signal.signal(signal.SIGTERM, the_handler)

    # Useful since this all will only be done once.
    if do_sysexit:
        the_handler.append(_sysexit_wrap)
    if add_previous:
        # Note that signal.SIG_DFL will be basically ignored.
        if callable(curhandler_term):
            the_handler.append(curhandler_term)
        if (callable(curhandler_int) and
                curhandler_int != curhandler_term):
            # (Note that same previous handler still can be called twice)
            the_handler.append(curhandler_int)
    if add_defaults:
        the_handler.append(_atexit_wrap)
    return the_handler



# coding: utf8
""" Things that are not quite okay to use in most cases.

Also, things that are useful in an ipython interactice shell.
"""

from __future__ import print_function, unicode_literals, absolute_import, division

from six.moves import builtins
from .base import PY_3
from . import madness_datadiff
from .madness_datadiff import *
from . import madness_oneliny
from .madness_oneliny import *
from . import madness_reprstuff
from .madness_reprstuff import *
from . import madness_stuffstuff
from .madness_stuffstuff import *


__all__ = (
    # __builtin__ hacks
    '_olt_into_builtin',
    '_into_builtin',
)
__all__ += madness_datadiff.__all__
__all__ += madness_oneliny.__all__
__all__ += madness_reprstuff.__all__
__all__ += madness_stuffstuff.__all__


if PY_3:
    from . import aio
    from .aio import *
    __all__ += aio.__all__


# # Builtin-madness # #


def _into_builtin(dct):
    """ Helper to put stuff (like the one-liner-helpers) into builtins """
    for key, val in dct.items():
        setattr(builtins, key, val)


# For _into_builtin
__all_stuff = locals()
__all_stuff_e = dict((key, globals().get(key)) for key in __all__)


try:
    # better pprint
    from IPython.lib.pretty import pprint, pretty
    __all_stuff.update(pprint=pprint, pretty=pretty, pformat=pretty)
    __all_stuff_e.update(pprint=pprint, pretty=pretty, pformat=pretty)
except ImportError as __e:
    print("What, no IPython?", __e)


# For explicit call:
def _olt_into_builtin():
    return _into_builtin(__all_stuff_e)


# For use as execfile()
if locals().get('__into_builtin'):
    _olt_into_builtin()


# Recommendation: put
#     from pyaux import madness
#     madness._olt_into_builtin()
# in the
#     ~/.config/ipython/profile_default/ipython_config.py



# coding: utf8
# pylint: disable=unused-import,wildcard-import,unused-wildcard-import,useless-object-inheritance

# NOTE: no modules imported here should import `decimal` (otherwise
#   `use_cdecimal` might become problematic for them)

from __future__ import absolute_import, print_function

import os
import sys

from copy import deepcopy
import errno
import functools
import itertools
from itertools import chain, repeat, islice
import json
import logging
import math
import re
import time
import traceback
import unicodedata


# Note: no `__all__` here so that for all things defined here, `from pyaux import ...` works too.

from .minisix import PY3, PY_3, text_type, unicode, izip, xrange, reraise

# Compat.
from . import ranges
from .ranges import *

from . import interpolate
from .interpolate import *

from . import iterables
from .iterables import *

from . import monkeys
from .monkeys import *
PY_35 = sys.version_info >= (3, 5)
PY_352 = sys.version_info >= (3, 5, 2)


def bubble(*args, **kwargs):
    """
    Prettified super():
    Calls `super(ThisClass, this_instance).this_method(...)`.

    Not super-performant but quite prettifying ("Performance is 5
    times worse than super() call").

    src:
    http://stackoverflow.com/questions/2706623/super-in-python-2-x-without-args/2706703#2706703
    """
    import inspect

    def find_class_by_code_object(back_self, method_name, code):
        for cls in inspect.getmro(type(back_self)):
            if method_name in cls.__dict__:
                method_fun = getattr(cls, method_name)
                if method_fun.im_func.func_code is code:
                    return cls
        raise Exception("Should not have ended here")

    frame = inspect.currentframe().f_back
    back_self = frame.f_locals['self']
    method_name = frame.f_code.co_name

    for _ in xrange(5):
        code = frame.f_code
        cls = find_class_by_code_object(back_self, method_name, code)
        if cls:
            super_ = super(cls, back_self)
            return getattr(super_, method_name)(*args, **kwargs)
        try:
            frame = frame.f_back
        except Exception:
            return None
    raise Exception("Went too far")


class dotdict(dict):
    """ A simple dict subclass with items also available over attributes """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            # A little less confusing way:
            _et, _ev, _tb = sys.exc_info()
            reraise(AttributeError, AttributeError(*_ev.args), _tb)

    def __setattr__(self, name, value):
        self[name] = value


# Compat alias
SmartDict = dotdict


_dotdictify_marker = object()


class dotdictify(dict):
    """ Recursive automatic doctdict thingy """

    def __init__(self, value=None):
        if value is None:
            pass
        elif isinstance(value, dict):
            for key in value:
                self.__setitem__(key, value[key])
        else:
            raise TypeError('expected a dict')

    def __setitem__(self, key, value):
        if isinstance(value, dict) and not isinstance(value, dotdictify):
            value = dotdictify(value)
        dict.__setitem__(self, key, value)

    def __getitem__(self, key):
        found = self.get(key, _dotdictify_marker)
        if found is _dotdictify_marker:
            found = dotdictify()
            dict.__setitem__(self, key, found)
        return found

    __setattr__ = __setitem__
    __getattr__ = __getitem__


def repr_call(args, kwargs):
    """ A helper function for pretty-printing a function call arguments """
    res = ', '.join("%r" % (val,) for val in args)
    if kwargs:
        if res:
            res += ', '
        res += ', '.join('%s=%r' % (key, val) for key, val in kwargs.items())
    return res


dbgs = {}   # global object for easier later access of dumped `__call__`s


def DebugPlug(name, mklogger=None):
    """ Create and return a recursive duck-object for plugging in
    place of other objects for debug purposes.

    :param name: name for tracking the object and its (child) attributes.

    :param mklogger: a function of `name` (str) that returns a new callable(msg:
      str) used for logging.  See code for an example.
    """
    # The construction with mkClass is for removing the need of
    #   `__getattr__`ing the name and logger.
    def mklogger_default(name):
        logger = logging.getLogger(name)
        return logger.debug

    if mklogger is None:
        mklogger = mklogger_default

    log = mklogger(name)

    class DebugPlugInternal(object):
        """ An actual internal class of the DebugPlug """

        def __call__(self, *ar, **kwa):
            log("called with (%s)", repr_call(ar, kwa))
            global dbgs  ## Just to note it
            dbgs.setdefault(name, {})
            dbgs[name]['__call__'] = (ar, kwa)

        def __getattr__(self, attname):
            namef = "%s.%s" % (name, attname)
            # Recursive!
            dpchild = DebugPlug(name=namef, mklogger=mklogger)
            # setattr(self, attname, dpchild)
            object.__setattr__(self, attname, dpchild)
            return dpchild

        def __setattr__(self, attname, value):
            log("setattr: %s = %r", attname, value)
            global dbgs
            dbgs.setdefault(name, {})
            dbgs[name][attname] = value
            return object.__setattr__(self, attname, value)

    return DebugPlugInternal()


# ###### dict-lazies

def dict_fget(D, k, d):
    """ dict_get(D, k, d) -> D[k] if k in D, else d().
      - a lazy-evaluated dict.get.
      (d is mandatory but can be None). """
    if k in D:
        return D[k]
    return d() if d is not None else d


def dict_fsetdefault(D, k, d):
    """ dict_fsetdefault(D, k, d) -> dict_fget(D, k, d), also set D[k]=d() if k not in D.
      - a lazy-evaluated dict.setdefault.
      (d is mandatory but can be None).  """
    # Can be `D[k] = dict_fget(D, k, d); return D[k]`, but let's micro-optimize.
    # NOTE: not going over 'keyerror' for the defaultdict or alike classes.
    if k in D:
        return D[k]
    v = d() if d is not None else d
    D[k] = v
    return v



def split_list(iterable, condition):
    """
    Split list items into `(matching, non_matching)` by `cond(item)` callable.
    """
    matching = []
    non_matching = []
    for item in iterable:
        if condition(item):
            matching.append(item)
        else:
            non_matching.append(item)
    return matching, non_matching


def dict_maybe_items(value):
    items = getattr(value, 'items', None)
    if items is not None:
        return items()
    return value


def split_dict(source, condition, cls=dict):
    source = dict_maybe_items(source)
    matching, nonmatching = split_list(
        source,
        lambda item: condition(item[0], item[1]))
    return cls(matching), cls(nonmatching)


# ...
# TODO: make a recurser that supports and recurses, maintains,
# optionally-replaces various classes such as tuple, defaultdict, and
# other containers. Possibly use pickle/deepcopy's logic.


def obj2dict(obj, add_type=False, add_instance=False, do_lists=True,
             dict_class=dotdict):
    """" Recursive obj -> obj.__dict__ """
    kwa = dict(
        add_type=add_type, add_instance=add_instance,
        do_lists=do_lists, dict_class=dict_class)

    if hasattr(obj, '__dict__'):
        res = dict_class()
        for key, val in obj.__dict__.items():
            res[key] = obj2dict(val, **kwa)
        if add_type:
            res['__class__'] = obj.__class__
        if add_instance:
            res['__instance__'] = obj
        return res

    # Recurse through other types too:
    # NOTE: There might be subclasses of these that would not be
    # processed here.
    if isinstance(obj, dict):
        return dict_class(
            (key, obj2dict(val, **kwa))
            for key, val in obj.items())

    if isinstance(obj, list):
        return [obj2dict(val, **kwa) for val in obj]

    return obj  # something else - return as-is.


def mk_logging_property(actual_name, logger_name='_log'):
    """
    Creates a property that logs the value and the caller in the
    setter, using logger under `self`'s logger_name, and stores the value
    under `actual_name` on `self`.
    """

    def do_get(self):
        return getattr(self, actual_name)

    def do_set(self, val):
        tb = traceback.extract_stack(limit=2)[0]
        # # or:
        # next((r.f_code.co_filename, r.f_lineno, r.f_code.co_name) for r in (sys._getframe(1),))
        # # that is,
        # r = sys._getframe(1)
        # co = r.f_code
        # co.co_filename, r.f_lineno, co.co_name
        setattr(self, actual_name, val)
        getattr(self, logger_name).debug(
            "%s set to %r from %s:%d, in %s",
            actual_name, val, tb[0], tb[1], tb[2])

    return property(do_get, do_set)


def sign(value):
    """
    Sign of value.

    One of the many possible ways to implement this.

    In [10]: sign(-10)
    Out[10]: -1

    In [11]: sign(0)
    Out[11]: 0

    In [12]: sign(10)
    Out[12]: 1
    """
    if value == 0:
        return 0
    if value > 0:
        return 1
    if value < 0:
        return -1
    raise ValueError("Inconsistent value")


# #######  "Human" sorting, advanced #######
# partially from quodlibet/quodlibet/util/__init__.py
# partially from comix/src/filehandler.py


def try_parse(val, fn=int):  # NOTE: generally similar to `.madness._try`
    """ 'try parse' (with fn) """
    try:
        return fn(val)
    except Exception:
        return val


# Note: not localized (i.e. always as dot for decimal separator)
_re_alphanum_f = re.compile(r'[0-9]+(?:\.[0-9]+)?|[^0-9]+')


def _split_numeric_f(string):
    return [
        try_parse(val, fn=float)
        for val in _re_alphanum_f.findall(string)]


# Or to avoid interpreting numbers as float:
_re_alphanum_int = re.compile(r'\d+|\D+')


def _split_numeric(string):
    return [
        try_parse(val, fn=int)
        for val in _re_alphanum_int.findall(string)]


# Primary function:
def human_sort_key(string, normalize=unicodedata.normalize, floats=True):
    """ Sorting key for 'human' sorting """
    string = to_text(string)
    string = normalize("NFD", string.lower())
    split_fn = _split_numeric_f if floats else _split_numeric
    return string and split_fn(string)


# ####### ...

class ThrottledCall(object):
    """ Decorator for throttling calls to some functions (e.g. logging).
    Defined as class for various custom attributes and methods.
    Attributes:
      `handle_skip`: function(self, *ar, **kwa) to call when a call is
        skipped. (default: return None)
    Methods: `call_something`
    """

    # _last_call_time = None
    _call_time_throttle = None
    _call_cnt = 0  # (kept accurate; but can become ineffectively large)
    _call_cnt_throttle = 0  # next _call_cnt to call at
    _call_val = object()  # (some unique value at start)

    def __init__(self, fn=None, sec_limit=None, cnt_limit=None):
        """
        :param fn: function to call (can be customized later).
        :param sec_limit: skip call if less than `sec_limit` seconds since the last call.
        :param cnt_limit: call only once each `cnt_limit` calls.
        """
        self.fn = fn
        # # mimickry, v2
        # self.__call__ = wraps(fn)(self.__call__)
        self.sec_limit = sec_limit
        self.cnt_limit = cnt_limit
        doc = "%s (throttled)" % (fn.__doc__,)

        self.__doc__ = doc
        self.__call__.__doc__ = doc
        _f = getattr(self.__call__, '__func__')
        if _f:
            _f.__doc__ = doc
        self.handle_skip = lambda self, *ar, **kwa: None

    def __call__(self, *ar, **kwa):
        return self.call_something(self.fn, *ar, **kwa)

    def call_something(self, fn, *ar, **kwa):
        """ Call some (other) function with the same throttling """
        now = time.time()
        # do_call = True
        # # NOTE: `throttle_cnt(throttle_sec(fn))` is emulated if both
        # # are set.
        if self.cnt_limit is not None:
            self._call_cnt += 1
            if self._call_cnt >= self._call_cnt_throttle:
                self._call_cnt_throttle += self.cnt_limit
            else:
                return self.handle_skip(self, *ar, **kwa)
        if self.sec_limit is not None:
            if now > self._call_time_throttle:
                self._call_time_throttle = now + self.sec_limit
            else:
                return self.handle_skip(self, *ar, **kwa)
        res = fn(*ar, **kwa)
        return res

    def call_value(self, val, fn, *ar, **kwa):
        """ Call if the value hasn't changed (applying the other
        throttling parameters as well). Contains undocumented feature
        """
        if self._call_val != val:
            self._call_val = val
            if fn is None:
                fn = self.fn
            return self.call_something(fn, *ar, **kwa)
        return None

    def __repr__(self):
        return "<throttled_call(%r)>" % (self.fn,)

    # # Optional: mimickry
    # def __repr__(self):
    #     return repr(self.fn)
    # def __getattr__(self, v):
    #     return getattr(self.fn, v)


@functools.wraps(ThrottledCall)
def throttled_call(*args, **kwargs):
    """ Wraps the supplied function with ThrottledCall (or generates a
    wrapper with the supplied parameters). """
    if args:
        func = args[0]
        args = args[1:]
        if callable(func):
            # mimickry, v3
            return functools.wraps(func)(ThrottledCall(func, *args, **kwargs))
        # else:  # supplied some arguments as positional?
        # TODO: make a warning
        args = (func,) + args

    return lambda func: functools.wraps(func)(ThrottledCall(func, *args, **kwargs))


class lazystr(object):
    """
    A simple class for lazy-computed processing into string,
    e.g. for use in logging.

    Example:

        log(13, "stuff is %r",
            lazystr(lambda: ', '.join(stuff)))

    Note: no caching. Use `memoize` if it is needed.
    """

    def __init__(self, fn):
        self.fn = fn

    def __str__(self):
        return str(self.fn())

    def __repr__(self):
        return repr(self.fn())


class LazyRepr(object):
    """
    Alternative of `lazystr` that does not do additional `repr()`ing, i.e. the
    `func` must return a string.
    """

    def __init__(self, func):
        self.func = func

    def __repr__(self):
        return to_str(self.func())

    def __str__(self):
        return to_str(self.func())

    def __unicode__(self):
        return to_text(self.func())


uniq_g = iterables.uniq  # Compat.


def uniq(lst, key=lambda v: v):  # pylint: disable=function-redefined
    """ RTFS """
    return list(iterables.uniq(lst, key=key))


list_uniq = uniq_g  # Compat.  # pylint: disable=invalid-name


# Helper for o_repr that displays '???'
class ReprObj(object):
    """ A class for inserting specific text in __repr__ outputs.  """

    def __init__(self, txt):
        self.txt = txt

    def __repr__(self):
        return self.txt


# _err_obj = type(b'ErrObj', (object,), dict(__repr__=lambda self: '???'))()
_err_obj = ReprObj('???')


# It is similar to using self.__dict__ in repr() but works over dir()
def o_repr_g(
        o, _colors=False, _colors256=False, _colorvs=None, _method=2,
        _private=False, _callable=False):
    """ Represent (most of) data on a python object in readable
    way. Useful default for a __repr__.
    WARN: does not handle recursive structures; use carefully.  """

    # TODO: handle recursive structures (similar to dict.__repr__)
    # TODO: process base types (dict / list / ...) in a special way

    def _color(x16, x256=None):
        if not _colors:
            return ''
        if x16 is None and x256 is None:
            return '\x1b[00m'
        if _colors256 and x256:
            return '\x1b[38;5;' + str(x256) + 'm'
        return '\x1b[0' + str(x16) + 'm'

    _colorvs_add = _colorvs
    _colorvs = dict(
        base=('1;37', '230'),  # Base (punctuation)  # white / pink-white
        clsn=('1;36', '123'),  # Class name  # bright cyan / cyan-white
        attr=('0;32', '120'),  # Attribute name  # dark-green / green-white
        func=('0;34', '75'),  # Function name  # blue / blue-cyan
        val=('0;37', '252'),  # Value data  # light-gray / light-light-gray
        clear=(None, None),
    )
    _colorvs.update(_colorvs_add or {})

    def _colorv(n):
        return _color(*_colorvs[n])

    yield _colorv("base")
    yield '<'
    yield _colorv("clsn")
    yield str(o.__class__.__name__)
    yield _colorv("base")
    yield '('

    if _method == 3:
        # V3: check type for properties
        o_type = type(o)

    first = True

    for n in sorted(dir(o)):

        # skip 'private' stuff
        if n.startswith('_') and not _private:
            continue

        if first:
            first = False
        else:
            yield _colorv("base")
            yield ', '

        keytype = "attr"
        has_val = True

        # V2: try but fail
        if _method == 2:
            try:
                v = getattr(o, n)
                if callable(v):  # skip functions (... and other callables)
                    if not _callable:
                        has_val = False
                    keytype = "func"
            except Exception:
                v = _err_obj

        if _method == 3:
            # V3: check type for properties
            v_m = getattr(o_type, n, None)
            if v_m is not None and isinstance(v_m, property):
                continue  # skip properties
            v = getattr(o, n)

            # skip functions (... and other callables)
            if callable(v):
                if not _callable:
                    has_val = False
                keytype = "func"

        yield _colorv(keytype)
        yield str(n)  # NOTE: some cases (e.g. functions) will remain just names

        if has_val:
            try:
                v = repr(v)
            except Exception:
                v = _err_obj
            yield _colorv("base")
            yield '='
            yield _colorv("val")
            yield v

    yield _colorv("base")
    yield ')>'
    yield _colorv("clear")


def o_repr(o, **kwa):
    return ''.join(o_repr_g(o, **kwa))


class OReprMixin(object):
    def __repr__(self):
        return o_repr(self)


def stdin_lines(strip_newlines=True):
    """ Iterate over stdin lines in a 'line-buffered' way """
    while True:
        try:
            line = sys.stdin.readline()
        except KeyboardInterrupt:
            # Generally, there's no point in dropping a traceback in a
            # script within an interrupted shell pipe.
            return
        if not line:
            # No more data to read (otherwise it would at least have an "\n")
            break
        # This might not be the case if the stream terminates with a non-newline at the end.
        if strip_newlines and line[-1] == '\n':
            line = line[:-1]
        yield line


def stdout_lines(gen, flush=True):
    """
    Send lines from a generator / iterable to stdout in a line-buffered way.

    Generally intended to work with text strings rather than bytestrings.
    """
    for line in gen:
        if not PY_3:
            line = to_bytes(line)
        sys.stdout.write("%s\n" % (line,))
        if flush:
            sys.stdout.flush()


def dict_merge(target, source, instancecheck=None, dictclass=dict,
               del_obj=object(), _copy=True, inplace=False):
    """ do update() on 'dict of dicts of di...' structure recursively.
    Also, see sources for details.
    NOTE: does not keep target's specific tree structure (forces source's)
    :param del_obj: allows for deletion of keys if the key in the `source` is set to this.

    >>> data = {}
    >>> data = dict_merge(data, {'open_folders': {'my_folder_a': False}})
    >>> data
    {'open_folders': {'my_folder_a': False}}
    >>> data = dict_merge(data, {'open_folders': {'my_folder_b': True}})
    >>> assert data == {'open_folders': {'my_folder_a': False, 'my_folder_b': True}}
    >>> _del = object()
    >>> data = dict_merge(data, {'open_folders': {'my_folder_b': _del}}, del_obj=_del)
    >>> assert data == {'open_folders': {'my_folder_a': False}}
    """
    if instancecheck is None:  # funhorrible ducktypings
        def instancecheck_default(iv):
            return hasattr(iv, 'items')

        instancecheck = instancecheck_default

    # Recursive parameters shorthand
    kwa = dict(instancecheck=instancecheck, dictclass=dictclass, del_obj=del_obj)

    if _copy and not inplace:  # 'both are default'
        target = deepcopy(target)

    for key, val in source.items():
        if val is del_obj:
            target.pop(key, None)
        elif instancecheck(val):  # (val -> source -> items())
            # NOTE: if target[key] wasn't a dict - it will be, now.
            target[key] = dict_merge(
                dict_fget(target, key, dictclass), val, **kwa)
        else:  # nowhere to recurse into - just replace
            # NOTE: if target[key] was a dict - it won't be, anymore.
            target[key] = val

    return target


def _sqrt(var):
    """ Duck-compatible sqrt(), needed to support classes like Decimal.

    Approximately the same as in the python3's `statistics`. """
    try:
        return var.sqrt()
    except AttributeError:
        return math.sqrt(var)


def chunks(lst, size):
    """ Yield successive chunks from lst. No padding.  """
    for idx in xrange(0, len(lst), size):
        yield lst[idx:idx + size]


def group(lst, cls=dict):
    """ RTFS.

    Similar to dict(MultiValueDict(lst).lists())

    >>> group([(1, 1), (2, 2), (1, 3)])
    {1: [1, 3], 2: [2]}
    """
    res = cls()
    for key, val in lst:
        try:
            group_list = res[key]
        except KeyError:
            res[key] = [val]
        else:
            group_list.append(val)
    return res


def group2(lst, key=lambda v: v[0]):
    """ RTFS.

    Not particularly better than `group((key(val), val) for val in lst)`.
    """
    res = {}
    for v in lst:
        res.setdefault(key(v), []).append(v)
    return res.items()


def mangle_items(items, include=None, exclude=None, add=None, replace=None, replace_inplace=None):
    """
    Functional-style dict editing core (working with a list of pairs).

    >>> items = [(1, 2), (3, 4), (5, 6), (7, 8)]
    >>> mangle_items(items, include=[3, 5], add=[(9, 10)], replace=[(5, 66)])
    [(3, 4), (9, 10), (5, 66)]
    >>> mangle_items(items, include=[3, 5], replace_inplace=[(5, 66)])
    [(3, 4), (5, 66)]
    >>> mangle_items(items, exclude=[3, 7], add=[(9, 10)])
    [(1, 2), (5, 6), (9, 10)]
    """
    include = set(include) if include is not None else None
    exclude = set(exclude) if exclude is not None else None

    if replace is not None:
        if isinstance(replace, dict):
            replace = replace.items()
        add = add + replace
        exclude = exclude if exclude is not None else set()
        exclude = exclude | set(key for key, val in replace)

    res = items
    if include is not None:
        res = [(key, val) for key, val in res if key in include]
    if replace_inplace is not None:
        if not isinstance(replace_inplace, dict):
            replace_inplace = dict(replace_inplace)
        res = [(key, replace_inplace.get(key, val)) for key, val in res]
    if exclude is not None:
        res = [(key, val) for key, val in res if key not in exclude]

    # Make sure we end up with a copy in any case:
    if res is items:
        res = list(items)

    # ... functional-style `update`.
    # Almost the `dict(input_dict, **add)`, but better.
    if add is not None:
        if isinstance(add, dict):
            add = add.items()
        res.extend(add)
    return res


def mangle_dict(input_dict, include=None, exclude=None, add=None, _return_list=False, dcls=dict):
    """ Functional-style dict editing """
    items = input_dict.items()
    res = mangle_items(items, include=include, exclude=exclude, add=add)
    if _return_list:
        return res
    return dcls(res)


filterdict = mangle_dict


def colorize(text, fmt, outfmt='term', lexer_kwa=None, formatter_kwa=None, **kwa):
    """ Convenience method for running pygments """
    from pygments import highlight, lexers, formatters
    _colorize_lexers = dict(
        yaml='YamlLexer', diff='DiffLexer',
        py='PythonLexer', py2='PythonLexer', py3='Python3Lexer',
    )
    _colorize_formatters = dict(
        term='TerminalFormatter', term256='Terminal256Formatter',
        html='HtmlFormatter',
    )
    fmt = _colorize_lexers.get(fmt, fmt)
    outfmt = _colorize_formatters.get(outfmt, outfmt)

    lexer_cls = getattr(lexers, fmt)
    formatter_cls = getattr(formatters, outfmt)
    return highlight(
        text, lexer_cls(**(lexer_kwa or {})),
        formatter_cls(**(formatter_kwa or {})))


def colorize_yaml(text, **kwa):
    """ Attempt to colorize the yaml text using pygments (for console
    output) """
    return colorize(text, 'yaml', **kwa)


def colorize_diff(text, **kwa):
    """ Attempt to colorize the [unified] diff text using pygments
    (for console output) """
    return colorize(text, 'diff', **kwa)


def _dict_hashable_1(dct):
    """ Simple non-recursive dict -> hashable """
    return tuple(sorted(dct.items()))

_dict_hashable = _dict_hashable_1


def configurable_wrapper(wrapper_func):
    """
    Wrap a wrapper to make it conveniently configurable.

    The wrapper_func should accept a positional argument
    `func_to_wrap` and optional keyword arguments
    (e.g. `some_func = wrapper_func(some_func, option1=123)`).

    The resulting function accepts optional keyword arguments and
    optional function `func_to_wrap`. If `func_to_wrap` is not
    specified, it returns a wrapper, otherwise it returns a wrapped
    function.

    Saves from the need to remember to write `@wrapper()`, making it
    possible to write both `@wrapper_func` and
    `@wrapper_func(option1=123)`.

    >>> @configurable_wrapper
    ... def wrap_stuff(func, **options):
    ...     def wrapped(*ar, **kwa):
    ...         print("wrapped", func, ar, kwa, "with", options)
    ...         return func(*ar, **kwa)
    ...     return wrapped
    ...
    >>> @wrap_stuff
    ... def some_func(*ar, **kwa):
    ...     print("some_func", ar, kwa)
    ...
    >>> some_func(1, 2, b=3)  #doctest: +ELLIPSIS
    wrapped <function some_func at 0x...> (1, 2) {'b': 3} with {}
    some_func (1, 2) {'b': 3}
    >>>
    >>>
    >>> @wrap_stuff(option1="value1")
    ... def another_func(*ar, **kwa):
    ...     print("another_func", ar, kwa)
    ...
    >>> another_func(4, 5, c=6)  #doctest: +ELLIPSIS
    wrapped <function another_func at 0x...> (4, 5) {'c': 6} with {'option1': 'value1'}
    another_func (4, 5) {'c': 6}
    """

    # TODO?: use `decorator` module here?

    @functools.wraps(wrapper_func)
    def configurable_wrapper_func(func_to_wrap=None, **options):
        def configured_wrapper_func(func_to_wrap):
            return wrapper_func(func_to_wrap, **options)

        # Cases of `@wrapper_func` and `wrapper_func(func, **options)`
        if func_to_wrap is not None:
            return configured_wrapper_func(func_to_wrap)

        # Case of `wrapper_func(**options)`
        return configured_wrapper_func

    return configurable_wrapper_func


def simple_memoize_argless(func):
    """
    A very simple memoizer that saves the first call result permanently
    (ignoring the argument values).
    """

    _cache = {}
    _sentinel = object()

    @functools.wraps(func)
    def simple_cached_wrapped(*args, **kwargs):
        result = _cache.get(None, _sentinel)
        if result is not _sentinel:
            return result

        result = func(*args, **kwargs)
        _cache[None] = result
        return result

    # Make the cache more easily accessible
    simple_cached_wrapped._cache = _cache
    return simple_cached_wrapped


# TODO?: some clear-all-global-memos method. By singleton and weakrefs.
class memoize(object):

    def __init__(
            self, fn, timelimit=None, single_value=False,
            force_kwarg='_memoize_force_new', timelimit_kwarg='_memoize_timelimit_override'):
        """
        ...

        :param fn: function to memoize.

        :param timelimit: seconds, float: consider the cached value invalid if
        it is older than that.

        :param single_value: keep only one value memoized, i.e. clear the cache
        on function call.
        """
        self.log = logging.getLogger("%s.%r" % (__name__, self))
        self.fn = fn
        self.mem = {}
        self.timelimit = timelimit
        self.single_value = single_value
        self.force_kwarg = force_kwarg
        self.timelimit_kwarg = timelimit_kwarg
        # Internal attribute, for `memoize_method`.
        self.skip_first_arg = False
        functools.update_wrapper(self, fn)

    def memoize_clear_mem(self):
        self.mem.clear()

    def __call__(self, *ar, **kwa):
        now = time.time()  # NOTE: before the call
        override = kwa.pop(self.force_kwarg, False)
        timelimit = kwa.pop(self.timelimit_kwarg, "_")
        if timelimit == "_":
            timelimit = self.timelimit
        # TODO?: cleanup obsolete keys here sometimes.
        try:
            if self.skip_first_arg:
                key = (ar[1:], _dict_hashable(kwa))
            else:
                key = (ar, _dict_hashable(kwa))
            # XXX/TODO: make `key` a weakref
            then, res = self.mem[key]
        except KeyError:
            pass
        except TypeError:  # e.g. unhashable args
            self.log.warning("memoize: Trying to memoize unhashable args %r, %r", ar, kwa)
            return self.fn(*ar, **kwa)
        else:
            if override:
                pass  # asked to ignore the cache
            elif timelimit is None or (now - then) < timelimit:
                return res  # valid cache
            else:
                pass  # no valid cache
        # KeyError or obsolete result.
        if self.single_value:
            self.memoize_clear_mem()
        res = self.fn(*ar, **kwa)
        self.mem[key] = (now, res)
        return res


@configurable_wrapper
def memoize_method(func, memo_attr=None, **cfg):
    """ `memoize` for a method, saving the cache on an instance
    attribute.

    :param memo_attr: name of the attribute to save the cache on.

    :param timelimit: see `memoize`.
    """

    if memo_attr is None:
        # Have to use func id because the func can be subclassed, and
        # will have the same name on the same instance despite
        # different contents. And figuring out the correct class name
        # at this point (like `self.__some_attr` does) is not viable.
        memo_attr = '_cached_%s_%x' % (func.__name__, id(func))

    @functools.wraps(func)
    def _memoized_method(self, *c_ar, **c_kwa):
        # Has to be done after the instantiation, to make the cache have the
        # same lifetime as the instance.
        cache = getattr(self, memo_attr, None)
        if cache is None:
            cache = memoize(func, **cfg)
            cache.skip_first_arg = True
            setattr(self, memo_attr, cache)

        return cache(self, *c_ar, **c_kwa)

    return _memoized_method


def memoized_property(*ar, **cfg):
    """
    Return a property attribute for new-style classes that only calls its getter on the first
    access. The result is stored and on subsequent accesses is returned, preventing the need to
    call the getter any more.

    >>> import time
    >>> class C(object):
    ...     load_name_count = 0
    ...     @memoized_property
    ...     def name(self):
    ...         ''' name's docstring '''
    ...         self.load_name_count += 1
    ...         return "the name"
    ...     @memoized_property(timelimit=0.0001)
    ...     def timelimited(self):
    ...         time.sleep(0.0005)
    ...         return self.load_name_count
    >>> c = C()
    >>> c.load_name_count
    0
    >>> c.timelimited
    0
    >>> c.name
    'the name'
    >>> c.load_name_count
    1
    >>> c.name
    'the name'
    >>> c.load_name_count
    1
    >>> c.timelimited
    1
    """
    func = None
    if ar and callable(ar[0]):
        func = ar[0]
        ar = ar[1:]

    memoize_method_configured = memoize_method(None, *ar, **cfg)

    def mk_property(a_func):
        return property(memoize_method_configured(a_func))

    if func is not None:
        return mk_property(func)

    return mk_property


memoize_property = memoized_property

def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def _dict_to_hashable_json(dct, dumps=json.dumps):
    """ ...

    NOTE: Not meant to be performant; tends towards collisions.

    Use `id(dct)` for a performant way with the reverse tendency.
    """
    return dumps(dct, skipkeys=True, default=id)


class hashabledict_st(dict):
    """ sorted-tuple based hashable subclass of `dict` """
    def __hash__(self):
        return hash(tuple(sorted(list(self.items()))))


def group3(items, to_hashable=_dict_to_hashable_json):
    """ Same as `group` but supports dicts as keys (and returns list
    of pairs) """
    annotated = [(to_hashable(key), key, val) for key, val in items]
    hashes = dict((keyhash, key) for keyhash, key, val in annotated)
    groups = group((keyhash, val) for keyhash, key, val in annotated).items()
    return [(hashes[keyhash], lst) for keyhash, lst in groups]


def to_bytes(value, default=None, encoding='utf-8', errors='strict'):
    if isinstance(value, bytes):
        return value
    if isinstance(value, text_type):
        return value.encode(encoding, errors)
    if default is not None:
        return default(value)
    return value


def to_text(value, default=None, encoding='utf-8', errors='strict'):
    if isinstance(value, text_type):
        return value
    if isinstance(value, bytes):
        return value.decode(encoding, errors)
    if default is not None:
        return default(value)
    return value


to_unicode = to_text  # legacy name  # pylint: disable=invalid-name


if PY_3:
    to_str = to_text  # pylint: disable=invalid-name
else:
    to_str = to_bytes  # pylint: disable=invalid-name


def import_module(name, package=None):
    """ ...

    django.utils.importlib.import_module without the package support
    """
    __import__(name)
    return sys.modules[name]


def import_func(func_path, _check_callable=True):
    """ Get an object (e.g. a function / callable) by import path.

    supports '<module>:<path>' notation as well as '<module>.<func_name>'.
    """
    # # Somewhat borrowed from django.core.handlers.base.BaseHandler.load_middleware
    # from django.utils.importlib import import_module

    _exc_cls = Exception

    if ':' in func_path:
        f_module, f_name = func_path.split(':', 1)
    else:
        try:
            f_module, f_name = func_path.rsplit('.', 1)
        except ValueError as exc:
            raise _exc_cls("func_path isn't a path", func_path, exc)

    f_name_parts = f_name.split('.')

    try:
        mod = import_module(f_module)
    except ImportError as exc:
        raise _exc_cls("func_path's module cannot be imported", func_path, exc)

    try:
        here = mod
        for f_name_part in f_name_parts:
            if not f_name_part:
                continue  ## allows for weirder things to be done.
            here = getattr(here, f_name_part)
        func = here
    except AttributeError as exc:
        raise _exc_cls("func_path's module does not have the specified func", func_path, exc)

    if _check_callable and not callable(func):
        raise _exc_cls("func does not seem to be a callable", func_path, func)

    return func


def dict_is_subset(
        smaller_obj,
        larger_obj,
        recurse_iterables=False,
        require_structure_match=True):
    """ Recursive check "smaller_dict's keys are subset of
    larger_dict's keys.

    NOTE: in practice, supports non-dict values at top.

    >>> value = {"a": 1, "b": [2, {"c": 3, "d": None}]}
    >>> dict_is_subset({}, value)
    True
    >>> dict_is_subset({"a": 1}, value)
    True
    >>> dict_is_subset({"a": 2}, value)
    False
    >>> dict_is_subset({"a": {"x": 4}}, value, require_structure_match=False)
    True
    >>> dict_is_subset({"b": [2]}, value, recurse_iterables=False)
    False
    >>> dict_is_subset({"b": [2]}, value, recurse_iterables=True, require_structure_match=True)
    False
    >>> dict_is_subset({"b": [2]}, value, recurse_iterables=True, require_structure_match=False)
    True
    >>> dict_is_subset({"b": [2, {}]}, value, recurse_iterables=True)
    True
    >>> dict_is_subset({"b": [2, {"c": 3}]}, value, recurse_iterables=True)
    True
    >>> dict_is_subset({"b": [2, {"c": 4}]}, value, recurse_iterables=True)
    False
    """
    kwa = dict(
        recurse_iterables=recurse_iterables,
        require_structure_match=require_structure_match,
    )
    if isinstance(smaller_obj, dict):
        if not isinstance(larger_obj, dict):
            return False if require_structure_match else True

        # Both are dicts.
        for key, val in smaller_obj.items():
            try:
                lval = larger_obj[key]
            except KeyError:
                return False
            # 'compare' the values whatever they are
            if not dict_is_subset(val, lval, **kwa):
                return False

        return True

    # else:
    if recurse_iterables and hasattr(smaller_obj, '__iter__'):
        if not hasattr(larger_obj, '__iter__'):
            return False if require_structure_match else True
        # smaller_value_iter, larger_value_iter
        svi = iter(smaller_obj)
        lvi = iter(larger_obj)
        for sval, lval in izip(svi, lvi):
            if not dict_is_subset(sval, lval, **kwa):
                return False
        if require_structure_match:
            if not iterator_is_over(svi) or not iterator_is_over(lvi):
                # One of the iterables was longer and thus was not
                # consumed entirely by the izip
                return False
        return True

    # else:
    # elif not dict or iterable:
    return smaller_obj == larger_obj


def find_files(
        in_dir, fname_re=None, older_than=None, skip_last=None,
        _prewalk=True, strip_dir=False, include_base=False):
    """ Return all full file paths under the directory `in_dir` whose
    *filenames* match the `fname_re` regexp (if not None) """
    now = time.time()
    walk = os.walk(in_dir)
    if _prewalk:
        walk = list(walk)
    for dir_name, _, file_list in walk:
        if fname_re is not None:
            file_list = (val for val in file_list if re.match(fname_re, val))

        # Annotate with full path:
        file_list = (
            (os.path.join(dir_name, file_name), file_name)
            for file_name in file_list)

        if strip_dir:
            # Strip the top dir from it
            file_list = (
                (slstrip(fpath, dir_name).lstrip('/'), fname)
                for fpath, fname in file_list)

        if older_than is not None:
            file_list = (
                (fpath, fname) for fpath, fname in file_list
                if now - os.path.getmtime(fpath) >= older_than)

        # Convenience shortcut
        if skip_last:
            file_list = sorted(list(file_list), key=lambda val: val[1])
            file_list = file_list[:-int(skip_last)]

        if not include_base:
            file_list = (fpath for fpath, fname in file_list)

        for res_val in file_list:
            yield res_val


@memoize
def get_requests_session():
    """
    Singleton with the common requests session (for maximal connection reuse).

    WARNING: Effectively deprecated; see `pyaux.req`.
    """
    import requests
    return requests.session()


def request(
        url, data=None,
        method=None, _w_method='post',
        # Conveniences for overriding:
        _extra_headers=None, _extra_params=None, _default_host=None,
        # NOTE: using JSON by default here.
        _dataser='json', timeout=5,
        _callinfo=True, session=True, _rfs=False, **kwa):
    """ requests.request wrapper with conveniences.

    :param session: default session if `True`, new session if `False`, or the specified session.
    :param _w_method: method to use for writing (if `data` or `files` are present).
    :param _dataser: data-dict serialization format (WARN: default is 'json'). 'json' or 'url'.
    :param _callinfo: add the caller file and line to the user-agent.

    WARNING: Effectively deprecated; see `pyaux.req`.
    """
    import requests
    from .minisix import urllib_parse
    log = logging.getLogger('request')

    # Different default, basically.
    kwa['timeout'] = timeout

    if url.startswith('/'):
        if _default_host is None:
            raise Exception("Must specify _default_host for host-relative URLs")
        if '://' not in _default_host:
            _default_host = 'http://%s' % _default_host
        url = urllib_parse.urljoin(_default_host, url)

    params = kwa.get('params') or {}
    if _extra_params:
        params.update(_extra_params)
    kwa['params'] = params

    headers = kwa.get('headers') or {}
    if _extra_headers:
        headers.update(_extra_headers)

    if _callinfo:
        if isinstance(_callinfo, tuple) and len(_callinfo) == 3:
            _cfile, _cline, _cfunc = _callinfo
        else:
            # TODO: custom function, extra_depth param.
            _cfile, _cline, _cfunc = logging.Logger('').findCaller()
        _prev_ua = headers.get('User-Agent') or requests.utils.default_user_agent()
        headers.setdefault('User-Agent', '%(ua)s, %(cfile)s:%(cline)s: %(cfunc)s' % dict(
            ua=_prev_ua, cfile=_cfile, cline=_cline, cfunc=_cfunc))

    is_writing = data is not None or kwa.get('files') is not None
    method = method if method is not None else (
        _w_method if is_writing else 'get')
    if method == 'get':
        # From requests.get:
        kwa.setdefault('allow_redirects', True)

    # Put them back
    kwa['headers'] = headers

    if session in (0, False, None):
        reqr = requests
    elif session in (1, True):
        reqr = get_requests_session()
    else:
        reqr = session

    if data:
        if isinstance(data, (bytes, unicode)):
            # Assume the data is already serialised
            pass
        elif _dataser == 'json':
            data = json.dumps(data)
            kwa.setdefault("headers", {}).setdefault("content-type", "application/json")
        elif _dataser in ('url', 'urlencode', None):
            pass  # pretty much the default in `requests`
        else:
            raise Exception("Unknown _dataser", _dataser)

        kwa['data'] = data
        # TODO?: log a piece of the data?

    # TODO?: put the params either entirely inthe url or entirely in
    # the dict (which shouldn't necessarily be a dict, although MVOD
    # is more convenient)?
    log.info("%s %s  params=%r", method.upper(), url, params)
    resp = reqr.request(method, url, **kwa)

    try:
        elapsed = '%.3fs' % (resp.elapsed.total_seconds(),)
    except Exception as exc:  # Just in case
        elapsed = '???(%s)' % (repr(exc)[:32],)
    # NOTE: assuming that the reponse content is never too large
    log.info(
        "Response: %s %s %s   %db in %s",
        resp.status_code, resp.request.method, resp.url, len(resp.content), elapsed)

    if _rfs:
        resp.raise_for_status()

    return resp  # The sufficiently convenient way


def exclogwrap(func=None, name=None, log=logging):
    """
    Wrap the function to exception-log its exceptions.

    Useful for e.g. send_robust signals.
    """


    def exclogwrap_configured(func):
        name_actual = name
        if name_actual is None and func is not None:
            name_actual = repr(func)

        @functools.wraps(func)
        def _wrapped(*ar, **kwa):
            try:
                return func(*ar, **kwa)
            except Exception as exc:
                log.exception("%r failed: %r", name_actual, exc)
                raise

        return _wrapped

    if func is not None:
        return exclogwrap_configured(func)

    return exclogwrap_configured


def repr_cut(some_str, length):
    if len(some_str) <= length:
        return some_str
    return some_str[:length] + (u'…' if isinstance(some_str, unicode) else '…')


def slstrip(self, substring):
    """ Strip a substring from the string at left side """
    if not self.startswith(substring):
        raise ValueError("Value %r does not start with substring %r" % (
            repr_cut(self, len(substring) * 2), substring))
    return self[len(substring):]


def get_env_flag(name, default=False, falses=('0',)):
    """
    Get a boolean flag from os.environ with support for `default`
    and some of the shell trickiness.

    To reset a value in shell, use `unset $name`.
    """
    try:
        value = os.environ[name]
    except KeyError:
        return default
    if value in falses:
        return False
    return value  # Still can be e.g. an empty string.


def current_frame(depth=1):
    """
    (from logging/__init__.py)
    """
    func = getattr(sys, '_getframe', None)
    if func is not None:
        return func(depth)
    # fallback; probably not relevant anymore.
    try:
        raise Exception
    except Exception:  # pylint: disable=broad-except
        frame = sys.exc_info()[2].tb_frame
        for _ in range(depth):
            frame = frame.f_back


def find_caller(extra_depth=1, skip_packages=()):
    """
    Find the stack frame of the caller so that we can note the source
    file name, line number and function name.

    Mostly a copypaste from `logging`.

    :param skip_packages: ...; example: `[getattr(logging, '_srcfile', None)]`.
    """
    cur_frame = current_frame(depth=2 + extra_depth)  # our caller, i.e. parent frame
    frame = cur_frame
    # On some versions of IronPython, currentframe() returns None if
    # IronPython isn't run with -X:Frames.
    result = "(unknown file)", 0, "(unknown function)"
    while hasattr(frame, "f_code"):
        codeobj = frame.f_code
        filename = os.path.normcase(codeobj.co_filename)
        # Additionally skip
        if any(filename.startswith(pkg) for pkg in skip_packages if pkg):
            frame = frame.f_back
            continue
        result = (codeobj.co_filename, frame.f_lineno, codeobj.co_name)
        break
    return result


try:
    monotonic_now = time.monotonic
except Exception:
    try:
        import monotonic  # pylint: disable=import-error
        monotonic_now = monotonic.monotonic
    except Exception:
        # No proper implementation available.
        monotonic_now = time.time


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# coding: utf8
"""
...
"""

from __future__ import division, absolute_import, print_function, unicode_literals

import socket


def get_constants(
        module, prefix,
        named=True, unprefixed=True, valued=True,
        lowercase=True,
        value_only=False):
    """ Mapping of constants by several ways to specify them """
    result = {}
    for name in dir(module):
        if not name.startswith(prefix):
            continue
        shortname = name[len(prefix):]
        value = getattr(module, name)
        if value_only:
            item = value
        else:
            item = dict(value=value, name=name, shortname=shortname)
        if named:
            result[name] = item
            if lowercase:
                result[name.lower()] = item
        if unprefixed:
            result[shortname] = item
            if lowercase:
                result[shortname.lower()] = item
        if valued:
            result[value] = item

    return result


SOCKET_FAMILIES = get_constants(socket, 'AF_')
SOCKET_TYPES = get_constants(socket, 'SOCK_')
SOCKET_PROTOS = get_constants(socket, 'IPPROTO_')
SOCKET_AI = get_constants(socket, 'AI_')


def gai_verbose(
        host,
        port=0, family=0, type='stream', proto='tcp',
        canonname=False, addrconfig=False, v4mapped=False, v4mapped_all=False,
        numerichost=False, numericserv=False, passive=False,
        flags=0):
    """

    :param addrconfig: return only address families (INET / INET6) that are configured on the current system.

    :param canonname: populate the "canonname" field ("official name of the host").

    :param v4mapped: return IPv4-mapped IPv6 addresses for hosts with no IPv6 addresses.

    :param v4mapped_all: return all IPv4 addresses as IPv4-mapped.

    :param numerichost: disallow network host address lookups.

    :param numericserv: disallow lookups of the service (`port`) names.

    :param passive: return addresses for binding sockets (rather than connecting).

    See also:
    http://man7.org/linux/man-pages/man3/getaddrinfo.3.html
    https://docs.python.org/3/library/socket.html
    """
    family_set = None
    if isinstance(family, (list, tuple, set)):
        family_set = set(
            SOCKET_FAMILIES[family_one]['value']
            for family_one in family)
        family = 0

    flags = flags or 0
    if addrconfig:
        flags = flags | socket.AI_ADDRCONFIG
    if canonname:
        flags = flags | socket.AI_CANONNAME
    if v4mapped or v4mapped_all:
        flags = flags | socket.AI_V4MAPPED
    if v4mapped_all:
        flags = flags | socket.AI_ALL
    if numerichost:
        flags = flags | socket.AI_NUMERICHOST
    if numericserv:
        flags = flags | socket.AI_NUMERICSERV
    if passive:
        flags = flags | socket.AI_PASSIVE

    responses = socket.getaddrinfo(
        host,
        port or 0,
        SOCKET_FAMILIES[family or 0]['value'],
        SOCKET_TYPES[type or 0]['value'],
        SOCKET_PROTOS[proto or 0]['value'],
        flags,
    )
    return gai_verbose_parse(responses, family_set=family_set, flags=flags)


def gai_verbose_parse(responses, family_set=None, flags=None):
    results = []
    flag_names = set()
    if flags is not None:
        flag_names = set(
            item['name']
            for num, item in SOCKET_AI.items()
            if flags & item['value']
        )

    for family, socktype, proto, canonname, sockaddr in responses:

        if family_set is not None and family not in family_set:
            continue
        # TODO: same for the rest.

        flow_info = None
        scope_id = None
        if len(sockaddr) == 2:
            address, port = sockaddr
        elif len(sockaddr) == 4:
            address, port, flow_info, scope_id = sockaddr
        else:
            raise Exception("Unexpected getaddrinfo sockaddr length", sockaddr)

        results.append(dict(
            address=address,
            port=port,
            flow_info=flow_info,
            scope_id=scope_id,

            canonname=canonname,
            family_name=SOCKET_FAMILIES[family]['name'],  # family.name,
            family=int(family),  # family.numerator,
            type_name=SOCKET_TYPES[socktype]['name'],  # socktype.name,
            type=int(socktype),  # socktype.numerator,
            proto_name=SOCKET_PROTOS[proto]['name'],
            proto=proto,

            sockaddr=sockaddr,
            flags=flags,
            flag_names=flag_names,
        ))
    return results


def _gai_example():
    hostname = 'ipv6-test.com'
    res = gai_verbose(
        hostname,
        family=('inet', 'inet6'),
        # canonname=True,
        # addrconfig=True,
        # v4mapped=True,
    )
    # res = pd.DataFrame(res, columns=res[0].keys() if res else None)
    return res


if __name__ == '__main__':
    print(_gai_example())



#!/usr/bin/env python
# encoding: utf-8
""" Module that adds coloring to the `logging` on `init()`.
src: http://stackoverflow.com/questions/384076/how-can-i-color-python-logging-output/1336640#1336640

TODO: make a handler/formatter with this instead.
"""

import logging


def add_coloring_to_emit_windows(fn):
        # add methods we need to the class
    def _out_handle(self):
        import ctypes
        return ctypes.windll.kernel32.GetStdHandle(self.STD_OUTPUT_HANDLE)
    out_handle = property(_out_handle)

    def _set_color(self, code):
        import ctypes
        # Constants from the Windows API
        self.STD_OUTPUT_HANDLE = -11
        hdl = ctypes.windll.kernel32.GetStdHandle(self.STD_OUTPUT_HANDLE)
        ctypes.windll.kernel32.SetConsoleTextAttribute(hdl, code)

    setattr(logging.StreamHandler, '_set_color', _set_color)

    def new(*args):
        FOREGROUND_BLUE = 0x0001  # text color contains blue.
        FOREGROUND_GREEN = 0x0002  # text color contains green.
        FOREGROUND_RED = 0x0004  # text color contains red.
        FOREGROUND_INTENSITY = 0x0008  # text color is intensified.
        FOREGROUND_WHITE = (
            FOREGROUND_BLUE | FOREGROUND_GREEN |
            FOREGROUND_RED)
       # winbase.h
        STD_INPUT_HANDLE = -10
        STD_OUTPUT_HANDLE = -11
        STD_ERROR_HANDLE = -12

        # wincon.h
        FOREGROUND_BLACK = 0x0000
        FOREGROUND_BLUE = 0x0001
        FOREGROUND_GREEN = 0x0002
        FOREGROUND_CYAN = 0x0003
        FOREGROUND_RED = 0x0004
        FOREGROUND_MAGENTA = 0x0005
        FOREGROUND_YELLOW = 0x0006
        FOREGROUND_GREY = 0x0007
        FOREGROUND_INTENSITY = 0x0008  # foreground color is intensified.

        BACKGROUND_BLACK = 0x0000
        BACKGROUND_BLUE = 0x0010
        BACKGROUND_GREEN = 0x0020
        BACKGROUND_CYAN = 0x0030
        BACKGROUND_RED = 0x0040
        BACKGROUND_MAGENTA = 0x0050
        BACKGROUND_YELLOW = 0x0060
        BACKGROUND_GREY = 0x0070
        BACKGROUND_INTENSITY = 0x0080  # background color is intensified.

        levelno = args[1].levelno
        if levelno >= 50:
            color = (BACKGROUND_YELLOW | FOREGROUND_RED |
                     FOREGROUND_INTENSITY | BACKGROUND_INTENSITY)
        elif levelno >= 40:
            color = FOREGROUND_RED | FOREGROUND_INTENSITY
        elif levelno >= 30:
            color = FOREGROUND_YELLOW | FOREGROUND_INTENSITY
        elif levelno >= 20:
            color = FOREGROUND_GREEN
        elif levelno >= 10:
            color = FOREGROUND_MAGENTA
        else:
            color = FOREGROUND_WHITE
        args[0]._set_color(color)

        ret = fn(*args)
        args[0]._set_color(FOREGROUND_WHITE)
        #print "after"
        return ret
    return new


def add_coloring_to_emit_ansi(fn):

    # add methods we need to the class
    def new(*args):
        record = args[1]
        levelno = record.levelno
        nocolor = '\x1b[0m'
        if levelno >= 50:
            color = '\x1b[31m'  # red
        elif levelno >= 40:
            color = '\x1b[31m'  # red
        elif levelno >= 30:
            color = '\x1b[33m'  # yellow
        elif levelno >= 20:
            color = '\x1b[32m'  # green
        elif levelno == 11:  # (useful for tmp-debug-messages)
            color = '\x1b[36m'  # cyan
        elif levelno >= 10:
            color = '\x1b[35m'  # pink
        else:
            color = nocolor  # normal
        record.msg = '%s%s%s' % (color, record.msg, nocolor)
        record.levelname = '%s%s%s' % (color, record.levelname, nocolor)
        # TODO?: color msg.name with hashfunc-color of `name` or `module`?
        return fn(*args)

    return new


def init():
    """ Monkey-patch to add color support to logging.StreamHandler """
    import platform
    if platform.system() == 'Windows':
        # Windows does not support ANSI escapes and we are using API calls
        # to set the console color
        logging.StreamHandler.emit = add_coloring_to_emit_windows(
            logging.StreamHandler.emit)
    else:
        # all non-Windows platforms are supporting ANSI escapes so we use them
        logging.StreamHandler.emit = add_coloring_to_emit_ansi(
            logging.StreamHandler.emit)
        # log = logging.getLogger()
        # log.addFilter(log_filter())
        # //hdlr = logging.StreamHandler()
        # //hdlr.setFormatter(formatter())


def test():
    """ Provide a simple test-demonstration """
    logging.basicConfig(level=logging.DEBUG)
    logging.debug("debug")
    logging.warning("a warning")
    logging.error("some error")
    logging.info("some info")


if __name__ == '__main__':
    test()



# coding: utf8
""" madstuff: repr stuff """

__all__ = (
    'GenReprWrap', 'GenReprWrapWrap',
)


class GenReprWrap(object):
    """ Generator proxy-wrapper that prints part of the child generator
    on __repr__ (saving it in a list).  """

    def __init__(self, gen, max_repr=20):
        self.gen = gen
        self.max_repr = max_repr
        self.cache_list = []
        self._probably_more = True

    def _make_cache(self):
        while len(self.cache_list) < self.max_repr:
            # NOTE: Not checking self._probably_more here, assuming
            #   that generator will continue to give out StopIteration
            #   many times.
            try:
                v = next(self.gen)
            except StopIteration:
                self._probably_more = False
                return self.cache_list
            self.cache_list.append(v)
        return self.cache_list

    def __repr__(self):
        cache = self._make_cache()
        res = []
        res.append('(')
        if cache:
            res.append(', '.join(repr(v) for v in cache))
        if self._probably_more:
            res.append(', ...')
        res.append(')')
        return ''.join(res)

    def next(self):
        try:  # Exhaust cache first.
            return self.cache_list.pop(0)
        except IndexError:
            return next(self.gen)

    def __iter__(self):
        return self


def GenReprWrapWrap(fn=None, **wrap_kwa):
    import functools

    def _wrap(w_fn):
        @functools.wraps(w_fn)
        def _wrapped(*ar, **kwa):
            res = fn(*ar, **kwa)
            if hasattr(res, '__iter__') and hasattr(res, 'next'):
                return GenReprWrap(res, **wrap_kwa)
            return res

        return _wrapped

    if fn is not None:
        return _wrap(fn)

    return _wrap
